from ._core import (
    DataRecord,
    DataRecordWrite,
    DomainModel,
    DomainModelCore,
    DomainModelWrite,
    DomainModelList,
    DomainRelationWrite,
    ResourcesWrite,
    ResourcesWriteResult,
)
from ._bid_method import (
    BidMethod,
    BidMethodApply,
    BidMethodApplyList,
    BidMethodFields,
    BidMethodList,
    BidMethodTextFields,
    BidMethodWrite,
    BidMethodWriteList,
)
from ._generator import (
    Generator,
    GeneratorApply,
    GeneratorApplyList,
    GeneratorFields,
    GeneratorList,
    GeneratorTextFields,
    GeneratorWrite,
    GeneratorWriteList,
)
from ._generator_efficiency_curve import (
    GeneratorEfficiencyCurve,
    GeneratorEfficiencyCurveApply,
    GeneratorEfficiencyCurveApplyList,
    GeneratorEfficiencyCurveFields,
    GeneratorEfficiencyCurveList,
    GeneratorEfficiencyCurveWrite,
    GeneratorEfficiencyCurveWriteList,
)
from ._plant import (
    Plant,
    PlantApply,
    PlantApplyList,
    PlantFields,
    PlantList,
    PlantTextFields,
    PlantWrite,
    PlantWriteList,
)
from ._price_area import (
    PriceArea,
    PriceAreaApply,
    PriceAreaApplyList,
    PriceAreaFields,
    PriceAreaList,
    PriceAreaTextFields,
    PriceAreaWrite,
    PriceAreaWriteList,
)
from ._reservoir import (
    Reservoir,
    ReservoirApply,
    ReservoirApplyList,
    ReservoirFields,
    ReservoirList,
    ReservoirTextFields,
    ReservoirWrite,
    ReservoirWriteList,
)
from ._turbine_efficiency_curve import (
    TurbineEfficiencyCurve,
    TurbineEfficiencyCurveApply,
    TurbineEfficiencyCurveApplyList,
    TurbineEfficiencyCurveFields,
    TurbineEfficiencyCurveList,
    TurbineEfficiencyCurveWrite,
    TurbineEfficiencyCurveWriteList,
)
from ._watercourse import (
    Watercourse,
    WatercourseApply,
    WatercourseApplyList,
    WatercourseFields,
    WatercourseList,
    WatercourseTextFields,
    WatercourseWrite,
    WatercourseWriteList,
)

Generator.model_rebuild()
GeneratorWrite.model_rebuild()
GeneratorApply.model_rebuild()
Plant.model_rebuild()
PlantWrite.model_rebuild()
PlantApply.model_rebuild()
PriceArea.model_rebuild()
PriceAreaWrite.model_rebuild()
PriceAreaApply.model_rebuild()
Watercourse.model_rebuild()
WatercourseWrite.model_rebuild()
WatercourseApply.model_rebuild()

__all__ = [
    "DataRecord",
    "DataRecordWrite",
    "ResourcesWrite",
    "DomainModel",
    "DomainModelCore",
    "DomainModelWrite",
    "DomainModelList",
    "DomainRelationWrite",
    "ResourcesWriteResult",
    "BidMethod",
    "BidMethodWrite",
    "BidMethodApply",
    "BidMethodList",
    "BidMethodWriteList",
    "BidMethodApplyList",
    "BidMethodFields",
    "BidMethodTextFields",
    "Generator",
    "GeneratorWrite",
    "GeneratorApply",
    "GeneratorList",
    "GeneratorWriteList",
    "GeneratorApplyList",
    "GeneratorFields",
    "GeneratorTextFields",
    "GeneratorEfficiencyCurve",
    "GeneratorEfficiencyCurveWrite",
    "GeneratorEfficiencyCurveApply",
    "GeneratorEfficiencyCurveList",
    "GeneratorEfficiencyCurveWriteList",
    "GeneratorEfficiencyCurveApplyList",
    "GeneratorEfficiencyCurveFields",
    "Plant",
    "PlantWrite",
    "PlantApply",
    "PlantList",
    "PlantWriteList",
    "PlantApplyList",
    "PlantFields",
    "PlantTextFields",
    "PriceArea",
    "PriceAreaWrite",
    "PriceAreaApply",
    "PriceAreaList",
    "PriceAreaWriteList",
    "PriceAreaApplyList",
    "PriceAreaFields",
    "PriceAreaTextFields",
    "Reservoir",
    "ReservoirWrite",
    "ReservoirApply",
    "ReservoirList",
    "ReservoirWriteList",
    "ReservoirApplyList",
    "ReservoirFields",
    "ReservoirTextFields",
    "TurbineEfficiencyCurve",
    "TurbineEfficiencyCurveWrite",
    "TurbineEfficiencyCurveApply",
    "TurbineEfficiencyCurveList",
    "TurbineEfficiencyCurveWriteList",
    "TurbineEfficiencyCurveApplyList",
    "TurbineEfficiencyCurveFields",
    "Watercourse",
    "WatercourseWrite",
    "WatercourseApply",
    "WatercourseList",
    "WatercourseWriteList",
    "WatercourseApplyList",
    "WatercourseFields",
    "WatercourseTextFields",
]



from __future__ import annotations

import datetime
from typing import TYPE_CHECKING

from cognite.client import data_modeling as dm, CogniteClient

from cognite.powerops.client._generated.day_ahead_bid.data_classes import (
    DomainModelCore,
    MultiScenarioMatrix,
    BidMethod,
)
from ._core import DEFAULT_QUERY_LIMIT, QueryBuilder, QueryStep, QueryAPI, T_DomainModelList, _create_edge_filter

if TYPE_CHECKING:
    from .alert_query import AlertQueryAPI
    from .shop_price_scenario_result_query import SHOPPriceScenarioResultQueryAPI


class MultiScenarioMatrixQueryAPI(QueryAPI[T_DomainModelList]):
    def __init__(
        self,
        client: CogniteClient,
        builder: QueryBuilder[T_DomainModelList],
        view_by_read_class: dict[type[DomainModelCore], dm.ViewId],
        filter_: dm.filters.Filter | None = None,
        limit: int = DEFAULT_QUERY_LIMIT,
    ):
        super().__init__(client, builder, view_by_read_class)

        self._builder.append(
            QueryStep(
                name=self._builder.next_name("multi_scenario_matrix"),
                expression=dm.query.NodeResultSetExpression(
                    from_=self._builder[-1].name if self._builder else None,
                    filter=filter_,
                ),
                select=dm.query.Select([dm.query.SourceSelector(self._view_by_read_class[MultiScenarioMatrix], ["*"])]),
                result_cls=MultiScenarioMatrix,
                max_retrieve_limit=limit,
            )
        )

    def alerts(
        self,
        external_id_prefix: str | None = None,
        space: str | list[str] | None = None,
        limit: int | None = DEFAULT_QUERY_LIMIT,
        retrieve_method: bool = False,
    ) -> AlertQueryAPI[T_DomainModelList]:
        """Query along the alert edges of the multi scenario matrix.

        Args:
            external_id_prefix: The prefix of the external ID to filter on.
            space: The space to filter on.
            limit: Maximum number of alert edges to return. Defaults to 25. Set to -1, float("inf") or None
                to return all items.
            retrieve_method: Whether to retrieve the method for each multi scenario matrix or not.

        Returns:
            AlertQueryAPI: The query API for the alert.
        """
        from .alert_query import AlertQueryAPI

        from_ = self._builder[-1].name

        edge_filter = _create_edge_filter(
            dm.DirectRelationReference("power-ops-types", "calculationIssue"),
            external_id_prefix=external_id_prefix,
            space=space,
        )
        self._builder.append(
            QueryStep(
                name=self._builder.next_name("alerts"),
                expression=dm.query.EdgeResultSetExpression(
                    filter=edge_filter,
                    from_=from_,
                    direction="outwards",
                ),
                select=dm.query.Select(),
                max_retrieve_limit=limit,
            )
        )
        if retrieve_method:
            self._query_append_method(from_)
        return AlertQueryAPI(self._client, self._builder, self._view_by_read_class, None, limit)

    def scenario_results(
        self,
        external_id_prefix: str | None = None,
        space: str | list[str] | None = None,
        limit: int | None = DEFAULT_QUERY_LIMIT,
        retrieve_method: bool = False,
    ) -> SHOPPriceScenarioResultQueryAPI[T_DomainModelList]:
        """Query along the scenario result edges of the multi scenario matrix.

        Args:
            external_id_prefix: The prefix of the external ID to filter on.
            space: The space to filter on.
            limit: Maximum number of scenario result edges to return. Defaults to 25. Set to -1, float("inf") or None
                to return all items.
            retrieve_method: Whether to retrieve the method for each multi scenario matrix or not.

        Returns:
            SHOPPriceScenarioResultQueryAPI: The query API for the shop price scenario result.
        """
        from .shop_price_scenario_result_query import SHOPPriceScenarioResultQueryAPI

        from_ = self._builder[-1].name

        edge_filter = _create_edge_filter(
            dm.DirectRelationReference("power-ops-types", "scenarioResult"),
            external_id_prefix=external_id_prefix,
            space=space,
        )
        self._builder.append(
            QueryStep(
                name=self._builder.next_name("scenario_results"),
                expression=dm.query.EdgeResultSetExpression(
                    filter=edge_filter,
                    from_=from_,
                    direction="outwards",
                ),
                select=dm.query.Select(),
                max_retrieve_limit=limit,
            )
        )
        if retrieve_method:
            self._query_append_method(from_)
        return SHOPPriceScenarioResultQueryAPI(self._client, self._builder, self._view_by_read_class, None, limit)

    def query(
        self,
        retrieve_method: bool = False,
    ) -> T_DomainModelList:
        """Execute query and return the result.

        Args:
            retrieve_method: Whether to retrieve the method for each multi scenario matrix or not.

        Returns:
            The list of the source nodes of the query.

        """
        from_ = self._builder[-1].name
        if retrieve_method:
            self._query_append_method(from_)
        return self._query()

    def _query_append_method(self, from_: str) -> None:
        view_id = self._view_by_read_class[BidMethod]
        self._builder.append(
            QueryStep(
                name=self._builder.next_name("method"),
                expression=dm.query.NodeResultSetExpression(
                    filter=dm.filters.HasData(views=[view_id]),
                    from_=from_,
                    through=self._view_by_read_class[MultiScenarioMatrix].as_property_ref("method"),
                    direction="outwards",
                ),
                select=dm.query.Select([dm.query.SourceSelector(view_id, ["*"])]),
                max_retrieve_limit=-1,
                result_cls=BidMethod,
            ),
        )



from __future__ import annotations

from collections.abc import Sequence
from typing import overload
import warnings

from cognite.client import CogniteClient
from cognite.client import data_modeling as dm
from cognite.client.data_classes.data_modeling.instances import InstanceAggregationResultList

from cognite.powerops.client._generated.v1.data_classes._core import DEFAULT_INSTANCE_SPACE
from cognite.powerops.client._generated.v1.data_classes import (
    DomainModelCore,
    DomainModelWrite,
    ResourcesWriteResult,
    SHOPTimeSeries,
    SHOPTimeSeriesWrite,
    SHOPTimeSeriesFields,
    SHOPTimeSeriesList,
    SHOPTimeSeriesWriteList,
    SHOPTimeSeriesTextFields,
)
from cognite.powerops.client._generated.v1.data_classes._shop_time_series import (
    _SHOPTIMESERIES_PROPERTIES_BY_FIELD,
    _create_shop_time_series_filter,
)
from ._core import (
    DEFAULT_LIMIT_READ,
    DEFAULT_QUERY_LIMIT,
    Aggregations,
    NodeAPI,
    SequenceNotStr,
    QueryStep,
    QueryBuilder,
)
from .shop_time_series_timeseries import SHOPTimeSeriesTimeseriesAPI
from .shop_time_series_query import SHOPTimeSeriesQueryAPI


class SHOPTimeSeriesAPI(NodeAPI[SHOPTimeSeries, SHOPTimeSeriesWrite, SHOPTimeSeriesList]):
    def __init__(self, client: CogniteClient, view_by_read_class: dict[type[DomainModelCore], dm.ViewId]):
        view_id = view_by_read_class[SHOPTimeSeries]
        super().__init__(
            client=client,
            sources=view_id,
            class_type=SHOPTimeSeries,
            class_list=SHOPTimeSeriesList,
            class_write_list=SHOPTimeSeriesWriteList,
            view_by_read_class=view_by_read_class,
        )
        self._view_id = view_id
        self.timeseries = SHOPTimeSeriesTimeseriesAPI(client, view_id)

    def __call__(
        self,
        object_type: str | list[str] | None = None,
        object_type_prefix: str | None = None,
        object_name: str | list[str] | None = None,
        object_name_prefix: str | None = None,
        attribute_name: str | list[str] | None = None,
        attribute_name_prefix: str | None = None,
        external_id_prefix: str | None = None,
        space: str | list[str] | None = None,
        limit: int | None = DEFAULT_QUERY_LIMIT,
        filter: dm.Filter | None = None,
    ) -> SHOPTimeSeriesQueryAPI[SHOPTimeSeriesList]:
        """Query starting at shop time series.

        Args:
            object_type: The object type to filter on.
            object_type_prefix: The prefix of the object type to filter on.
            object_name: The object name to filter on.
            object_name_prefix: The prefix of the object name to filter on.
            attribute_name: The attribute name to filter on.
            attribute_name_prefix: The prefix of the attribute name to filter on.
            external_id_prefix: The prefix of the external ID to filter on.
            space: The space to filter on.
            limit: Maximum number of shop time series to return. Defaults to 25. Set to -1, float("inf") or None to return all items.
            filter: (Advanced) If the filtering available in the above is not sufficient, you can write your own filtering which will be ANDed with the filter above.

        Returns:
            A query API for shop time series.

        """
        has_data = dm.filters.HasData(views=[self._view_id])
        filter_ = _create_shop_time_series_filter(
            self._view_id,
            object_type,
            object_type_prefix,
            object_name,
            object_name_prefix,
            attribute_name,
            attribute_name_prefix,
            external_id_prefix,
            space,
            (filter and dm.filters.And(filter, has_data)) or has_data,
        )
        builder = QueryBuilder(SHOPTimeSeriesList)
        return SHOPTimeSeriesQueryAPI(self._client, builder, self._view_by_read_class, filter_, limit)

    def apply(
        self,
        shop_time_series: SHOPTimeSeriesWrite | Sequence[SHOPTimeSeriesWrite],
        replace: bool = False,
        write_none: bool = False,
    ) -> ResourcesWriteResult:
        """Add or update (upsert) shop time series.

        Args:
            shop_time_series: Shop time series or sequence of shop time series to upsert.
            replace (bool): How do we behave when a property value exists? Do we replace all matching and existing values with the supplied values (true)?
                Or should we merge in new values for properties together with the existing values (false)? Note: This setting applies for all nodes or edges specified in the ingestion call.
            write_none (bool): This method, will by default, skip properties that are set to None. However, if you want to set properties to None,
                you can set this parameter to True. Note this only applies to properties that are nullable.
        Returns:
            Created instance(s), i.e., nodes, edges, and time series.

        Examples:

            Create a new shop_time_series:

                >>> from cognite.powerops.client._generated.v1 import PowerOpsModelsV1Client
                >>> from cognite.powerops.client._generated.v1.data_classes import SHOPTimeSeriesWrite
                >>> client = PowerOpsModelsV1Client()
                >>> shop_time_series = SHOPTimeSeriesWrite(external_id="my_shop_time_series", ...)
                >>> result = client.shop_time_series.apply(shop_time_series)

        """
        warnings.warn(
            "The .apply method is deprecated and will be removed in v1.0. "
            "Please use the .upsert method on the client instead. This means instead of "
            "`my_client.shop_time_series.apply(my_items)` please use `my_client.upsert(my_items)`."
            "The motivation is that all apply methods are the same, and having one apply method per API "
            " class encourages users to create items in small batches, which is inefficient."
            "In addition, .upsert method is more descriptive of what the method does.",
            UserWarning,
            stacklevel=2,
        )
        return self._apply(shop_time_series, replace, write_none)

    def delete(
        self, external_id: str | SequenceNotStr[str], space: str = DEFAULT_INSTANCE_SPACE
    ) -> dm.InstancesDeleteResult:
        """Delete one or more shop time series.

        Args:
            external_id: External id of the shop time series to delete.
            space: The space where all the shop time series are located.

        Returns:
            The instance(s), i.e., nodes and edges which has been deleted. Empty list if nothing was deleted.

        Examples:

            Delete shop_time_series by id:

                >>> from cognite.powerops.client._generated.v1 import PowerOpsModelsV1Client
                >>> client = PowerOpsModelsV1Client()
                >>> client.shop_time_series.delete("my_shop_time_series")
        """
        warnings.warn(
            "The .delete method is deprecated and will be removed in v1.0. "
            "Please use the .delete method on the client instead. This means instead of "
            "`my_client.shop_time_series.delete(my_ids)` please use `my_client.delete(my_ids)`."
            "The motivation is that all delete methods are the same, and having one delete method per API "
            " class encourages users to delete items in small batches, which is inefficient.",
            UserWarning,
            stacklevel=2,
        )
        return self._delete(external_id, space)

    @overload
    def retrieve(self, external_id: str, space: str = DEFAULT_INSTANCE_SPACE) -> SHOPTimeSeries | None: ...

    @overload
    def retrieve(self, external_id: SequenceNotStr[str], space: str = DEFAULT_INSTANCE_SPACE) -> SHOPTimeSeriesList: ...

    def retrieve(
        self, external_id: str | SequenceNotStr[str], space: str = DEFAULT_INSTANCE_SPACE
    ) -> SHOPTimeSeries | SHOPTimeSeriesList | None:
        """Retrieve one or more shop time series by id(s).

        Args:
            external_id: External id or list of external ids of the shop time series.
            space: The space where all the shop time series are located.

        Returns:
            The requested shop time series.

        Examples:

            Retrieve shop_time_series by id:

                >>> from cognite.powerops.client._generated.v1 import PowerOpsModelsV1Client
                >>> client = PowerOpsModelsV1Client()
                >>> shop_time_series = client.shop_time_series.retrieve("my_shop_time_series")

        """
        return self._retrieve(external_id, space)

    def search(
        self,
        query: str,
        properties: SHOPTimeSeriesTextFields | Sequence[SHOPTimeSeriesTextFields] | None = None,
        object_type: str | list[str] | None = None,
        object_type_prefix: str | None = None,
        object_name: str | list[str] | None = None,
        object_name_prefix: str | None = None,
        attribute_name: str | list[str] | None = None,
        attribute_name_prefix: str | None = None,
        external_id_prefix: str | None = None,
        space: str | list[str] | None = None,
        limit: int | None = DEFAULT_LIMIT_READ,
        filter: dm.Filter | None = None,
    ) -> SHOPTimeSeriesList:
        """Search shop time series

        Args:
            query: The search query,
            properties: The property to search, if nothing is passed all text fields will be searched.
            object_type: The object type to filter on.
            object_type_prefix: The prefix of the object type to filter on.
            object_name: The object name to filter on.
            object_name_prefix: The prefix of the object name to filter on.
            attribute_name: The attribute name to filter on.
            attribute_name_prefix: The prefix of the attribute name to filter on.
            external_id_prefix: The prefix of the external ID to filter on.
            space: The space to filter on.
            limit: Maximum number of shop time series to return. Defaults to 25. Set to -1, float("inf") or None to return all items.
            filter: (Advanced) If the filtering available in the above is not sufficient, you can write your own filtering which will be ANDed with the filter above.

        Returns:
            Search results shop time series matching the query.

        Examples:

           Search for 'my_shop_time_series' in all text properties:

                >>> from cognite.powerops.client._generated.v1 import PowerOpsModelsV1Client
                >>> client = PowerOpsModelsV1Client()
                >>> shop_time_series_list = client.shop_time_series.search('my_shop_time_series')

        """
        filter_ = _create_shop_time_series_filter(
            self._view_id,
            object_type,
            object_type_prefix,
            object_name,
            object_name_prefix,
            attribute_name,
            attribute_name_prefix,
            external_id_prefix,
            space,
            filter,
        )
        return self._search(self._view_id, query, _SHOPTIMESERIES_PROPERTIES_BY_FIELD, properties, filter_, limit)

    @overload
    def aggregate(
        self,
        aggregations: (
            Aggregations
            | dm.aggregations.MetricAggregation
            | Sequence[Aggregations]
            | Sequence[dm.aggregations.MetricAggregation]
        ),
        property: SHOPTimeSeriesFields | Sequence[SHOPTimeSeriesFields] | None = None,
        group_by: None = None,
        query: str | None = None,
        search_properties: SHOPTimeSeriesTextFields | Sequence[SHOPTimeSeriesTextFields] | None = None,
        object_type: str | list[str] | None = None,
        object_type_prefix: str | None = None,
        object_name: str | list[str] | None = None,
        object_name_prefix: str | None = None,
        attribute_name: str | list[str] | None = None,
        attribute_name_prefix: str | None = None,
        external_id_prefix: str | None = None,
        space: str | list[str] | None = None,
        limit: int | None = DEFAULT_LIMIT_READ,
        filter: dm.Filter | None = None,
    ) -> list[dm.aggregations.AggregatedNumberedValue]: ...

    @overload
    def aggregate(
        self,
        aggregations: (
            Aggregations
            | dm.aggregations.MetricAggregation
            | Sequence[Aggregations]
            | Sequence[dm.aggregations.MetricAggregation]
        ),
        property: SHOPTimeSeriesFields | Sequence[SHOPTimeSeriesFields] | None = None,
        group_by: SHOPTimeSeriesFields | Sequence[SHOPTimeSeriesFields] = None,
        query: str | None = None,
        search_properties: SHOPTimeSeriesTextFields | Sequence[SHOPTimeSeriesTextFields] | None = None,
        object_type: str | list[str] | None = None,
        object_type_prefix: str | None = None,
        object_name: str | list[str] | None = None,
        object_name_prefix: str | None = None,
        attribute_name: str | list[str] | None = None,
        attribute_name_prefix: str | None = None,
        external_id_prefix: str | None = None,
        space: str | list[str] | None = None,
        limit: int | None = DEFAULT_LIMIT_READ,
        filter: dm.Filter | None = None,
    ) -> InstanceAggregationResultList: ...

    def aggregate(
        self,
        aggregate: (
            Aggregations
            | dm.aggregations.MetricAggregation
            | Sequence[Aggregations]
            | Sequence[dm.aggregations.MetricAggregation]
        ),
        property: SHOPTimeSeriesFields | Sequence[SHOPTimeSeriesFields] | None = None,
        group_by: SHOPTimeSeriesFields | Sequence[SHOPTimeSeriesFields] | None = None,
        query: str | None = None,
        search_property: SHOPTimeSeriesTextFields | Sequence[SHOPTimeSeriesTextFields] | None = None,
        object_type: str | list[str] | None = None,
        object_type_prefix: str | None = None,
        object_name: str | list[str] | None = None,
        object_name_prefix: str | None = None,
        attribute_name: str | list[str] | None = None,
        attribute_name_prefix: str | None = None,
        external_id_prefix: str | None = None,
        space: str | list[str] | None = None,
        limit: int | None = DEFAULT_LIMIT_READ,
        filter: dm.Filter | None = None,
    ) -> list[dm.aggregations.AggregatedNumberedValue] | InstanceAggregationResultList:
        """Aggregate data across shop time series

        Args:
            aggregate: The aggregation to perform.
            property: The property to perform aggregation on.
            group_by: The property to group by when doing the aggregation.
            query: The query to search for in the text field.
            search_property: The text field to search in.
            object_type: The object type to filter on.
            object_type_prefix: The prefix of the object type to filter on.
            object_name: The object name to filter on.
            object_name_prefix: The prefix of the object name to filter on.
            attribute_name: The attribute name to filter on.
            attribute_name_prefix: The prefix of the attribute name to filter on.
            external_id_prefix: The prefix of the external ID to filter on.
            space: The space to filter on.
            limit: Maximum number of shop time series to return. Defaults to 25. Set to -1, float("inf") or None to return all items.
            filter: (Advanced) If the filtering available in the above is not sufficient, you can write your own filtering which will be ANDed with the filter above.

        Returns:
            Aggregation results.

        Examples:

            Count shop time series in space `my_space`:

                >>> from cognite.powerops.client._generated.v1 import PowerOpsModelsV1Client
                >>> client = PowerOpsModelsV1Client()
                >>> result = client.shop_time_series.aggregate("count", space="my_space")

        """

        filter_ = _create_shop_time_series_filter(
            self._view_id,
            object_type,
            object_type_prefix,
            object_name,
            object_name_prefix,
            attribute_name,
            attribute_name_prefix,
            external_id_prefix,
            space,
            filter,
        )
        return self._aggregate(
            self._view_id,
            aggregate,
            _SHOPTIMESERIES_PROPERTIES_BY_FIELD,
            property,
            group_by,
            query,
            search_property,
            limit,
            filter_,
        )

    def histogram(
        self,
        property: SHOPTimeSeriesFields,
        interval: float,
        query: str | None = None,
        search_property: SHOPTimeSeriesTextFields | Sequence[SHOPTimeSeriesTextFields] | None = None,
        object_type: str | list[str] | None = None,
        object_type_prefix: str | None = None,
        object_name: str | list[str] | None = None,
        object_name_prefix: str | None = None,
        attribute_name: str | list[str] | None = None,
        attribute_name_prefix: str | None = None,
        external_id_prefix: str | None = None,
        space: str | list[str] | None = None,
        limit: int | None = DEFAULT_LIMIT_READ,
        filter: dm.Filter | None = None,
    ) -> dm.aggregations.HistogramValue:
        """Produces histograms for shop time series

        Args:
            property: The property to use as the value in the histogram.
            interval: The interval to use for the histogram bins.
            query: The query to search for in the text field.
            search_property: The text field to search in.
            object_type: The object type to filter on.
            object_type_prefix: The prefix of the object type to filter on.
            object_name: The object name to filter on.
            object_name_prefix: The prefix of the object name to filter on.
            attribute_name: The attribute name to filter on.
            attribute_name_prefix: The prefix of the attribute name to filter on.
            external_id_prefix: The prefix of the external ID to filter on.
            space: The space to filter on.
            limit: Maximum number of shop time series to return. Defaults to 25. Set to -1, float("inf") or None to return all items.
            filter: (Advanced) If the filtering available in the above is not sufficient, you can write your own filtering which will be ANDed with the filter above.

        Returns:
            Bucketed histogram results.

        """
        filter_ = _create_shop_time_series_filter(
            self._view_id,
            object_type,
            object_type_prefix,
            object_name,
            object_name_prefix,
            attribute_name,
            attribute_name_prefix,
            external_id_prefix,
            space,
            filter,
        )
        return self._histogram(
            self._view_id,
            property,
            interval,
            _SHOPTIMESERIES_PROPERTIES_BY_FIELD,
            query,
            search_property,
            limit,
            filter_,
        )

    def list(
        self,
        object_type: str | list[str] | None = None,
        object_type_prefix: str | None = None,
        object_name: str | list[str] | None = None,
        object_name_prefix: str | None = None,
        attribute_name: str | list[str] | None = None,
        attribute_name_prefix: str | None = None,
        external_id_prefix: str | None = None,
        space: str | list[str] | None = None,
        limit: int | None = DEFAULT_LIMIT_READ,
        filter: dm.Filter | None = None,
    ) -> SHOPTimeSeriesList:
        """List/filter shop time series

        Args:
            object_type: The object type to filter on.
            object_type_prefix: The prefix of the object type to filter on.
            object_name: The object name to filter on.
            object_name_prefix: The prefix of the object name to filter on.
            attribute_name: The attribute name to filter on.
            attribute_name_prefix: The prefix of the attribute name to filter on.
            external_id_prefix: The prefix of the external ID to filter on.
            space: The space to filter on.
            limit: Maximum number of shop time series to return. Defaults to 25. Set to -1, float("inf") or None to return all items.
            filter: (Advanced) If the filtering available in the above is not sufficient, you can write your own filtering which will be ANDed with the filter above.

        Returns:
            List of requested shop time series

        Examples:

            List shop time series and limit to 5:

                >>> from cognite.powerops.client._generated.v1 import PowerOpsModelsV1Client
                >>> client = PowerOpsModelsV1Client()
                >>> shop_time_series_list = client.shop_time_series.list(limit=5)

        """
        filter_ = _create_shop_time_series_filter(
            self._view_id,
            object_type,
            object_type_prefix,
            object_name,
            object_name_prefix,
            attribute_name,
            attribute_name_prefix,
            external_id_prefix,
            space,
            filter,
        )
        return self._list(limit=limit, filter=filter_)



from __future__ import annotations

import datetime
from collections.abc import Sequence
from concurrent.futures import ThreadPoolExecutor
from typing import Callable, Literal, Optional, Union, overload
from urllib.parse import urlparse

import arrow
import requests
from cognite.client import CogniteClient
from cognite.client.data_classes import UserProfile, filters
from cognite.client.data_classes.events import EventSort
from cognite.client.exceptions import CogniteAPIError

from cognite.powerops.cdf_labels import RelationshipLabel
from cognite.powerops.client.shop.shop_run_filter import SHOPRunFilter
from cognite.powerops.utils.cdf.resource_creation import simple_relationship

from .data_classes.dayahead_trigger import Case, PrerunFileMetadata, ShopPreRunFile
from .shop_case import SHOPCase, SHOPFileReference, SHOPFileType
from .shop_run import SHOPRun, ShopRunEvent, SHOPRunList
from .utils import new_external_id

DEFAULT_READ_LIMIT = 25

_APICallableForSHOPRunT = Callable[[Union[SHOPRun, Sequence[SHOPRun]]], Union[SHOPRun, Sequence[SHOPRun]]]


class SHOPRunAPI:
    def __init__(self, client: CogniteClient, dataset_id: int, cogshop_version: str = ""):
        self._cdf = client
        self._dataset_id = dataset_id
        self.cogshop_version = cogshop_version
        self._CONCURRENT_CALLS = 5

    def _get_shop_prerun_files(self, file_external_ids: list[str]) -> list[ShopPreRunFile]:
        prerun_files = self._cdf.files.retrieve_multiple(external_ids=file_external_ids)
        return [ShopPreRunFile.load_from_metadata(file) for file in prerun_files]

    def trigger_case(self, case: Case, shop_version: str) -> tuple[list, list[SHOPRun]]:
        """
        Trigger a collection of shop runs related to one Case (also referred to as watercourse).
        For each ShopCase the prerun file will be used to trigger a shop run event in cdf,
        and used to trigger the CogShop container.
        Prerun files must exist in CDF.
        """
        shop_events = []
        now = datetime.datetime.now(datetime.timezone.utc).replace(microsecond=0)
        plants_per_case = []
        pre_runs = self._get_shop_prerun_files(case.pre_run_file_external_ids)
        for shop_run in pre_runs:
            plants_per_case.extend(shop_run.plants)
            shop_events.append(
                SHOPRun(
                    external_id=new_external_id(now=now),
                    watercourse=case.case_name,
                    start=now,
                    end=None,
                    data_set_id=self._dataset_id,
                    _case_file_external_id=shop_run.pre_run_external_id,
                    _shop_files=(
                        [SHOPFileReference(external_id=case.commands_file, file_type=SHOPFileType.ASCII.value)]
                        if case.commands_file
                        else []
                    ),
                    shop_version=shop_version,
                    plants=PrerunFileMetadata._plants_delimiter.join(shop_run.plants),
                    price_scenario=shop_run.price_scenario,
                    _client=self._cdf,
                    source="DayaheadTrigger",
                )
            )
        self._cdf.events.create([new_event.as_cdf_event() for new_event in shop_events])

        with ThreadPoolExecutor(max_workers=self._CONCURRENT_CALLS) as executor:
            futures = [executor.submit(self._trigger_shop_container, shop_event) for shop_event in shop_events]

            for future in futures:
                if exc := future.exception():
                    raise (exc)

        return list(set(plants_per_case)), shop_events

    def trigger_single_casefile(self, case: SHOPCase, source: str | None = None) -> SHOPRun:
        """
        Trigger SHOP for a given case file.

        Args:
            case: SHOPCase instance, contains data for the main case file and references to additional SHOP input files.
            source: The source of the SHOP trigger. If nothing is passed, the method will
                try to detect the service principal of the current user.

        Returns:
            The new SHOP run created.
        """
        if not case.watercourse:
            raise ValueError("Set watercourse on case before triggering.")

        user: UserProfile | None
        try:
            user = self._cdf.iam.user_profiles.me()
        except CogniteAPIError as e:
            if e.code == 404:
                user = None
            else:
                raise

        if source is None:
            source = user.user_identifier if user else "manual"
        display_name = user.display_name if user else "Unknown"
        now = datetime.datetime.now(datetime.timezone.utc)

        case_file_meta = self._cdf.files.upload_bytes(
            # On Windows machines, some bytes can get lost in the encoding process
            # when uploading a string to CDF. This is a workaround.
            content=case.yaml.encode("utf-8"),
            name=f"Manual Case by {display_name}",
            mime_type="application/yaml",
            external_id=new_external_id(prefix="cog_shop_manual", now=now),
            metadata={"shop:type": "cog_shop_case", "user": display_name},
            data_set_id=self._dataset_id,
            source=source,
        )

        shop_run = SHOPRun(
            external_id=new_external_id(now=now),
            watercourse=case.watercourse,
            data_set_id=self._dataset_id,
            start=now,
            end=None,
            _case_file_external_id=case_file_meta.external_id,
            _shop_files=case.shop_files,
            shop_version=case.shop_version,
            _client=self._cdf,
            source=source,
        )
        event = shop_run.as_cdf_event()
        self._cdf.events.create(shop_run.as_cdf_event())
        relationships = [
            simple_relationship(source=event, target=case_file_meta, label_external_id=RelationshipLabel.CASE_FILE)
        ]

        for shop_file_reference in case.shop_files:
            cdf_file = shop_file_reference.as_cdf_file_metadata()
            relationships.append(
                simple_relationship(source=event, target=cdf_file, label_external_id=RelationshipLabel.EXTRA_FILE)
            )
        self._cdf.relationships.create(relationships)

        self._trigger_shop_container(shop_run)
        return shop_run

    def trigger_shop_run(self, shop_run: SHOPRun) -> SHOPRun:
        self._create(shop_run)
        self._trigger_shop_container(shop_run)
        return shop_run

    def _trigger_shop_container(self, shop_run: SHOPRun):
        def auth(r: requests.PreparedRequest) -> requests.PreparedRequest:
            auth_header_name, auth_header_value = self._cdf._config.credentials.authorization_header()
            r.headers[auth_header_name] = auth_header_value
            return r

        response = requests.post(
            url=self._shop_url(),
            json={"shopEventExternalId": shop_run.external_id, "cogShopVersion": self.cogshop_version},
            auth=auth,
        )
        response.raise_for_status()

    def _shop_url(self) -> str:
        project = self._cdf.config.project

        cluster = urlparse(self._cdf.config.base_url).netloc.split(".", 1)[0]

        environment = project.split("-")[-1]
        if environment in {"dev", "staging"}:
            stage = ".staging"
        elif environment == "prod":
            stage = ""
        else:
            raise ValueError(f"Can't detect prod/staging from project name: {project!r}")

        return f"https://power-ops-api{stage}.{cluster}.cognite.ai/{project}/run-shop"

    def _load_cdf_event_shop_runs(
        self,
        extra_filters: Optional[list[filters.Filter]] = None,
        limit: int = DEFAULT_READ_LIMIT,
        event_sort: EventSort = None,
    ) -> SHOPRunList:
        is_type = filters.Equals("type", ShopRunEvent.event_type)
        events = self._cdf.events.filter(filters.And(is_type, *extra_filters), limit=limit, sort=event_sort)
        return SHOPRunList.load(events)

    @overload
    def retrieve(self, external_id: str) -> SHOPRun | None: ...

    @overload
    def retrieve(self, external_id: Sequence[str]) -> SHOPRunList: ...

    def retrieve(
        self, external_id: str | Sequence[str], ignore_unknown_ids: bool = True
    ) -> SHOPRun | SHOPRunList | None:
        """
        Retrieve a SHOP run.

        Args:
            external_id: The external id(s) of the SHOP run(s) to retrieve.
            ignore_unknown_ids: Whether to ignore unknown ids or raise an error.

        Returns:
            The SHOP run(s). None if it is a single external id, and it is not found.
        """
        if isinstance(external_id, str):
            event = self._cdf.events.retrieve(external_id=external_id)
            return None if event is None else SHOPRun.load(event)
        elif isinstance(external_id, Sequence):
            return SHOPRunList.load(
                self._cdf.events.retrieve_multiple(external_ids=external_id, ignore_unknown_ids=ignore_unknown_ids)
            )
        else:
            raise TypeError(f"Invalid type {type(external_id)} for external_id.")

    def retrieve_latest(
        self,
        watercourse: str | list[str] | None = None,
        source: str | list[str] | None = None,
        start_after: str | arrow.Arrow | datetime.datetime | None = None,
        start_before: str | arrow.Arrow | datetime.datetime | None = None,
        end_after: str | arrow.Arrow | datetime.datetime | None = None,
        end_before: str | arrow.Arrow | datetime.datetime | None = None,
        latest_by: Literal[
            "start_time",
            "end_time",
            "created_time",
            "last_updated_time",
        ] = "last_updated_time",
    ) -> SHOPRun | None:
        """
        Retrieves the the latest shop run given the filters.

         Args:
            watercourse: The watercourse(s) to filter on.
            source: The source(s) to filter on.
            start_after: The start time after which the SHOP run must have started.
            start_before: The start time before which the SHOP run must have started.
            end_after: The end time after which the SHOP run must have ended.
            end_before: The end time before which the SHOP run must have ended.
            latest_by: The property to retrieve by . The most reliable is `last_updated_time` and `created_time`.
        Returns:
            The most recent SHOP run, None if nothing matched the filter
        """
        extra_filters = SHOPRunFilter(
            watercourse=watercourse,
            source=source,
            start_after=start_after,
            start_before=start_before,
            end_after=end_after,
            end_before=end_before,
        ).get_filters()
        event_sort = EventSort(property=latest_by, order="desc")

        shop_runs = self._load_cdf_event_shop_runs(extra_filters=extra_filters, limit=1, event_sort=event_sort)
        if len(shop_runs) == 0:
            return None

        return shop_runs[0]

    def list(
        self,
        watercourse: str | list[str] | None = None,
        source: str | list[str] | None = None,
        start_after: str | arrow.Arrow | datetime.datetime | None = None,
        start_before: str | arrow.Arrow | datetime.datetime | None = None,
        end_after: str | arrow.Arrow | datetime.datetime | None = None,
        end_before: str | arrow.Arrow | datetime.datetime | None = None,
        limit: int = DEFAULT_READ_LIMIT,
    ) -> SHOPRunList:
        """List the filtered SHOP runs.
        Provide time stamps strings as YYYY-MM-DD) or otherwise something parsable by `arrow.get()` can parse.

        Args:
            watercourse: The watercourse(s) to filter on.
            source: The source(s) to filter on. WARNING: Most
            start_after: The start time after which the SHOP run must have started.
            start_before: The start time before which the SHOP run must have started.
            end_after: The end time after which the SHOP run must have ended.
            end_before: The end time before which the SHOP run must have ended.
            limit: The maximum number of SHOP runs to return.

        Returns:
            A list of SHOP runs.
        """

        extra_filters = SHOPRunFilter(
            watercourse=watercourse,
            source=source,
            start_after=start_after,
            start_before=start_before,
            end_after=end_after,
            end_before=end_before,
        ).get_filters()
        return self._load_cdf_event_shop_runs(extra_filters=extra_filters, limit=limit)

    def versions(self, py_version: str = "3.9") -> list[str]:
        """List the available version of SHOP remotely  in CDF. Does not include versions in CogShop Image.
        SHOP releases should have the following format:
        'SHOP-${{VERSION}}-pyshop-python{py_version}.linux.zip'
        """
        files = self._cdf.files.list(metadata={"shop:type": "shop-release"}, limit=-1)
        remote_versions = []
        for file in files:
            name_parts = file.name.split("-")
            if f"python{py_version}" in name_parts:
                remote_versions.append("-".join(name_parts[:2]))
        return remote_versions

    @overload
    def _create(self, shop_run: SHOPRun | Sequence[SHOPRun]) -> SHOPRun | Sequence[SHOPRun]: ...

    @overload
    def _create(self, shop_run: Sequence[SHOPRun]) -> Sequence[SHOPRun]: ...

    def _create(self, shop_run: SHOPRun | Sequence[SHOPRun]) -> SHOPRun | Sequence[SHOPRun]:
        api_method = self._wrap_event_api(self._cdf.events.create)
        return api_method(shop_run)

    @overload
    def _update(self, shop_run: SHOPRun | Sequence[SHOPRun]) -> SHOPRun | Sequence[SHOPRun]: ...

    @overload
    def _update(self, shop_run: Sequence[SHOPRun]) -> Sequence[SHOPRun]: ...

    def _update(self, shop_run: SHOPRun | Sequence[SHOPRun]) -> SHOPRun | Sequence[SHOPRun]:
        api_method = self._wrap_event_api(self._cdf.events.update)
        return api_method(shop_run)

    @overload
    def _upsert(self, shop_run: SHOPRun | Sequence[SHOPRun]) -> SHOPRun | Sequence[SHOPRun]: ...

    @overload
    def _upsert(self, shop_run: Sequence[SHOPRun]) -> Sequence[SHOPRun]: ...

    def _upsert(self, shop_run: SHOPRun | Sequence[SHOPRun]) -> SHOPRun | Sequence[SHOPRun]:
        api_method = self._wrap_event_api(self._cdf.events.update)
        return api_method(shop_run)

    def _wrap_event_api(self, api_method: Callable) -> _APICallableForSHOPRunT:
        def wrapped(shop_run: SHOPRun | Sequence[SHOPRun]) -> SHOPRun | Sequence[SHOPRun]:
            is_single = isinstance(shop_run, SHOPRun)
            shop_runs = [shop_run] if is_single else shop_run
            new_events = api_method([shop_run.as_cdf_event() for shop_run in shop_runs])
            loaded_shop_runs = [SHOPRun.load(new_event) for new_event in new_events]
            return loaded_shop_runs[0] if is_single else loaded_shop_runs

        return wrapped



from __future__ import annotations

import logging

from cognite.client import CogniteClient

from cognite.powerops.resync.diff import ModelDifference
from cognite.powerops.resync.models.base import AssetModel, Model  # type: ignore[attr-defined]

logger = logging.getLogger(__name__)


def _clean_relationships(client: CogniteClient, differences: ModelDifference, new_model: Model):
    if isinstance(new_model, AssetModel):
        not_create = _find_relationship_with_missing_time_series_target(client, new_model)
        relationship_diff = next((d for d in differences if d.field_name == "relationships"), None)
        if relationship_diff:
            relationship_diff.added = [r for r in relationship_diff.added if r.external_id not in not_create]
            relationship_diff.changed = [c for c in relationship_diff.changed if c.new.external_id not in not_create]


def _find_relationship_with_missing_time_series_target(client: CogniteClient, asset_model: AssetModel) -> set[str]:
    """Validates that all relationships in the collection have targets that exist in CDF"""
    time_series = asset_model.timeseries()
    # retrieve_multiple fails if no time series are provided
    if not time_series:
        return set()

    existing_time_series = client.time_series.retrieve_multiple(
        external_ids=list(set(time_series.as_external_ids())), ignore_unknown_ids=True
    )
    existing_timeseries_ids = {ts.external_id: ts for ts in existing_time_series}
    missing_timeseries = {t.external_id for t in time_series if t.external_id not in existing_timeseries_ids}

    relationships = asset_model.relationships()
    to_delete = {
        r.external_id
        for r in relationships
        if r.target_type
        and r.target_type.casefold() == "timeseries"
        and r.target_external_id in missing_timeseries
        and r.external_id
    }
    if to_delete:
        logger.warning(
            f"There are {len(to_delete)} relationships in {asset_model.model_name} that have targets "
            f"that do not exist in CDF. Examples {list(to_delete)[:3]}. These relationships will not be created.",
        )
    return to_delete



"""
This package contains the v2 version of the models in the resync package. All of these models
are based on Data Modeling in CDF.

* Benchmarking - This is the data model version of the Benchmarking part of the MarketModel in v1.
* CogSHOP - This is the updated version of the CogShop1Asset in v1.
"""

from .production_dm import PowerAssetModelDM

__all__ = ["PowerAssetModelDM"]



from __future__ import annotations

import warnings
from typing import TYPE_CHECKING, Any, Literal, Optional, Union

from cognite.client import data_modeling as dm
from pydantic import Field

from ._core import (
    DEFAULT_INSTANCE_SPACE,
    DataRecordWrite,
    DomainModel,
    DomainModelCore,
    DomainModelWrite,
    DomainModelWriteList,
    DomainModelList,
    DomainRelationWrite,
    ResourcesWrite,
)

if TYPE_CHECKING:
    from ._alert import Alert, AlertWrite
    from ._bid_matrix import BidMatrix, BidMatrixWrite
    from ._partial_post_processing_input import PartialPostProcessingInput, PartialPostProcessingInputWrite


__all__ = [
    "PartialPostProcessingOutput",
    "PartialPostProcessingOutputWrite",
    "PartialPostProcessingOutputApply",
    "PartialPostProcessingOutputList",
    "PartialPostProcessingOutputWriteList",
    "PartialPostProcessingOutputApplyList",
    "PartialPostProcessingOutputFields",
    "PartialPostProcessingOutputTextFields",
]


PartialPostProcessingOutputTextFields = Literal["process_id", "function_name", "function_call_id"]
PartialPostProcessingOutputFields = Literal["process_id", "process_step", "function_name", "function_call_id"]

_PARTIALPOSTPROCESSINGOUTPUT_PROPERTIES_BY_FIELD = {
    "process_id": "processId",
    "process_step": "processStep",
    "function_name": "functionName",
    "function_call_id": "functionCallId",
}


class PartialPostProcessingOutput(DomainModel):
    """This represents the reading version of partial post processing output.

    It is used to when data is retrieved from CDF.

    Args:
        space: The space where the node is located.
        external_id: The external id of the partial post processing output.
        data_record: The data record of the partial post processing output node.
        process_id: The process associated with the function execution
        process_step: This is the step in the process.
        function_name: The name of the function
        function_call_id: The function call id
        alerts: An array of calculation level Alerts.
        input_: The input field.
        partial_matrices: The processed partial bid matrices that are used to calculate the total bid matrix.
    """

    space: str = DEFAULT_INSTANCE_SPACE
    node_type: Union[dm.DirectRelationReference, None] = dm.DirectRelationReference(
        "sp_powerops_types", "PartialPostProcessingOutput"
    )
    process_id: str = Field(alias="processId")
    process_step: int = Field(alias="processStep")
    function_name: str = Field(alias="functionName")
    function_call_id: str = Field(alias="functionCallId")
    alerts: Union[list[Alert], list[str], None] = Field(default=None, repr=False)
    input_: Union[PartialPostProcessingInput, str, dm.NodeId, None] = Field(None, repr=False, alias="input")
    partial_matrices: Union[list[BidMatrix], list[str], None] = Field(default=None, repr=False, alias="partialMatrices")

    def as_write(self) -> PartialPostProcessingOutputWrite:
        """Convert this read version of partial post processing output to the writing version."""
        return PartialPostProcessingOutputWrite(
            space=self.space,
            external_id=self.external_id,
            data_record=DataRecordWrite(existing_version=self.data_record.version),
            process_id=self.process_id,
            process_step=self.process_step,
            function_name=self.function_name,
            function_call_id=self.function_call_id,
            alerts=[alert.as_write() if isinstance(alert, DomainModel) else alert for alert in self.alerts or []],
            input_=self.input_.as_write() if isinstance(self.input_, DomainModel) else self.input_,
            partial_matrices=[
                partial_matrice.as_write() if isinstance(partial_matrice, DomainModel) else partial_matrice
                for partial_matrice in self.partial_matrices or []
            ],
        )

    def as_apply(self) -> PartialPostProcessingOutputWrite:
        """Convert this read version of partial post processing output to the writing version."""
        warnings.warn(
            "as_apply is deprecated and will be removed in v1.0. Use as_write instead.",
            UserWarning,
            stacklevel=2,
        )
        return self.as_write()


class PartialPostProcessingOutputWrite(DomainModelWrite):
    """This represents the writing version of partial post processing output.

    It is used to when data is sent to CDF.

    Args:
        space: The space where the node is located.
        external_id: The external id of the partial post processing output.
        data_record: The data record of the partial post processing output node.
        process_id: The process associated with the function execution
        process_step: This is the step in the process.
        function_name: The name of the function
        function_call_id: The function call id
        alerts: An array of calculation level Alerts.
        input_: The input field.
        partial_matrices: The processed partial bid matrices that are used to calculate the total bid matrix.
    """

    space: str = DEFAULT_INSTANCE_SPACE
    node_type: Union[dm.DirectRelationReference, None] = dm.DirectRelationReference(
        "sp_powerops_types", "PartialPostProcessingOutput"
    )
    process_id: str = Field(alias="processId")
    process_step: int = Field(alias="processStep")
    function_name: str = Field(alias="functionName")
    function_call_id: str = Field(alias="functionCallId")
    alerts: Union[list[AlertWrite], list[str], None] = Field(default=None, repr=False)
    input_: Union[PartialPostProcessingInputWrite, str, dm.NodeId, None] = Field(None, repr=False, alias="input")
    partial_matrices: Union[list[BidMatrixWrite], list[str], None] = Field(
        default=None, repr=False, alias="partialMatrices"
    )

    def _to_instances_write(
        self,
        cache: set[tuple[str, str]],
        view_by_read_class: dict[type[DomainModelCore], dm.ViewId] | None,
        write_none: bool = False,
    ) -> ResourcesWrite:
        resources = ResourcesWrite()
        if self.as_tuple_id() in cache:
            return resources

        write_view = (view_by_read_class or {}).get(
            PartialPostProcessingOutput, dm.ViewId("sp_powerops_models", "PartialPostProcessingOutput", "1")
        )

        properties: dict[str, Any] = {}

        if self.process_id is not None:
            properties["processId"] = self.process_id

        if self.process_step is not None:
            properties["processStep"] = self.process_step

        if self.function_name is not None:
            properties["functionName"] = self.function_name

        if self.function_call_id is not None:
            properties["functionCallId"] = self.function_call_id

        if self.input_ is not None:
            properties["input"] = {
                "space": self.space if isinstance(self.input_, str) else self.input_.space,
                "externalId": self.input_ if isinstance(self.input_, str) else self.input_.external_id,
            }

        if properties:
            this_node = dm.NodeApply(
                space=self.space,
                external_id=self.external_id,
                existing_version=self.data_record.existing_version,
                type=self.node_type,
                sources=[
                    dm.NodeOrEdgeData(
                        source=write_view,
                        properties=properties,
                    )
                ],
            )
            resources.nodes.append(this_node)
            cache.add(self.as_tuple_id())

        edge_type = dm.DirectRelationReference("sp_powerops_types", "calculationIssue")
        for alert in self.alerts or []:
            other_resources = DomainRelationWrite.from_edge_to_resources(
                cache, start_node=self, end_node=alert, edge_type=edge_type, view_by_read_class=view_by_read_class
            )
            resources.extend(other_resources)

        edge_type = dm.DirectRelationReference("sp_powerops_types", "BidMatrix")
        for partial_matrice in self.partial_matrices or []:
            other_resources = DomainRelationWrite.from_edge_to_resources(
                cache,
                start_node=self,
                end_node=partial_matrice,
                edge_type=edge_type,
                view_by_read_class=view_by_read_class,
            )
            resources.extend(other_resources)

        if isinstance(self.input_, DomainModelWrite):
            other_resources = self.input_._to_instances_write(cache, view_by_read_class)
            resources.extend(other_resources)

        return resources


class PartialPostProcessingOutputApply(PartialPostProcessingOutputWrite):
    def __new__(cls, *args, **kwargs) -> PartialPostProcessingOutputApply:
        warnings.warn(
            "PartialPostProcessingOutputApply is deprecated and will be removed in v1.0. Use PartialPostProcessingOutputWrite instead."
            "The motivation for this change is that Write is a more descriptive name for the writing version of the"
            "PartialPostProcessingOutput.",
            UserWarning,
            stacklevel=2,
        )
        return super().__new__(cls)


class PartialPostProcessingOutputList(DomainModelList[PartialPostProcessingOutput]):
    """List of partial post processing outputs in the read version."""

    _INSTANCE = PartialPostProcessingOutput

    def as_write(self) -> PartialPostProcessingOutputWriteList:
        """Convert these read versions of partial post processing output to the writing versions."""
        return PartialPostProcessingOutputWriteList([node.as_write() for node in self.data])

    def as_apply(self) -> PartialPostProcessingOutputWriteList:
        """Convert these read versions of primitive nullable to the writing versions."""
        warnings.warn(
            "as_apply is deprecated and will be removed in v1.0. Use as_write instead.",
            UserWarning,
            stacklevel=2,
        )
        return self.as_write()


class PartialPostProcessingOutputWriteList(DomainModelWriteList[PartialPostProcessingOutputWrite]):
    """List of partial post processing outputs in the writing version."""

    _INSTANCE = PartialPostProcessingOutputWrite


class PartialPostProcessingOutputApplyList(PartialPostProcessingOutputWriteList): ...


def _create_partial_post_processing_output_filter(
    view_id: dm.ViewId,
    process_id: str | list[str] | None = None,
    process_id_prefix: str | None = None,
    min_process_step: int | None = None,
    max_process_step: int | None = None,
    function_name: str | list[str] | None = None,
    function_name_prefix: str | None = None,
    function_call_id: str | list[str] | None = None,
    function_call_id_prefix: str | None = None,
    input_: str | tuple[str, str] | list[str] | list[tuple[str, str]] | None = None,
    external_id_prefix: str | None = None,
    space: str | list[str] | None = None,
    filter: dm.Filter | None = None,
) -> dm.Filter | None:
    filters = []
    if isinstance(process_id, str):
        filters.append(dm.filters.Equals(view_id.as_property_ref("processId"), value=process_id))
    if process_id and isinstance(process_id, list):
        filters.append(dm.filters.In(view_id.as_property_ref("processId"), values=process_id))
    if process_id_prefix is not None:
        filters.append(dm.filters.Prefix(view_id.as_property_ref("processId"), value=process_id_prefix))
    if min_process_step is not None or max_process_step is not None:
        filters.append(
            dm.filters.Range(view_id.as_property_ref("processStep"), gte=min_process_step, lte=max_process_step)
        )
    if isinstance(function_name, str):
        filters.append(dm.filters.Equals(view_id.as_property_ref("functionName"), value=function_name))
    if function_name and isinstance(function_name, list):
        filters.append(dm.filters.In(view_id.as_property_ref("functionName"), values=function_name))
    if function_name_prefix is not None:
        filters.append(dm.filters.Prefix(view_id.as_property_ref("functionName"), value=function_name_prefix))
    if isinstance(function_call_id, str):
        filters.append(dm.filters.Equals(view_id.as_property_ref("functionCallId"), value=function_call_id))
    if function_call_id and isinstance(function_call_id, list):
        filters.append(dm.filters.In(view_id.as_property_ref("functionCallId"), values=function_call_id))
    if function_call_id_prefix is not None:
        filters.append(dm.filters.Prefix(view_id.as_property_ref("functionCallId"), value=function_call_id_prefix))
    if input_ and isinstance(input_, str):
        filters.append(
            dm.filters.Equals(
                view_id.as_property_ref("input"), value={"space": DEFAULT_INSTANCE_SPACE, "externalId": input_}
            )
        )
    if input_ and isinstance(input_, tuple):
        filters.append(
            dm.filters.Equals(view_id.as_property_ref("input"), value={"space": input_[0], "externalId": input_[1]})
        )
    if input_ and isinstance(input_, list) and isinstance(input_[0], str):
        filters.append(
            dm.filters.In(
                view_id.as_property_ref("input"),
                values=[{"space": DEFAULT_INSTANCE_SPACE, "externalId": item} for item in input_],
            )
        )
    if input_ and isinstance(input_, list) and isinstance(input_[0], tuple):
        filters.append(
            dm.filters.In(
                view_id.as_property_ref("input"), values=[{"space": item[0], "externalId": item[1]} for item in input_]
            )
        )
    if external_id_prefix is not None:
        filters.append(dm.filters.Prefix(["node", "externalId"], value=external_id_prefix))
    if isinstance(space, str):
        filters.append(dm.filters.Equals(["node", "space"], value=space))
    if space and isinstance(space, list):
        filters.append(dm.filters.In(["node", "space"], values=space))
    if filter:
        filters.append(filter)
    return dm.filters.And(*filters) if filters else None



from .to_cogshop_model import to_cogshop_asset_model
from .to_market_model import to_market_asset_model
from .to_production_model import to_production_model

__all__ = ["to_production_model", "to_market_asset_model", "to_cogshop_asset_model"]



from typing import Union

from cognite.client import ClientConfig, CogniteClient
from cognite.client.credentials import OAuthClientCredentials, OAuthDeviceCode

from cognite.powerops.utils.cdf._settings import CogniteSettings, Settings


def get_client_config(settings: CogniteSettings) -> ClientConfig:
    credentials: Union[OAuthClientCredentials, OAuthDeviceCode]
    if settings.login_flow == "client_credentials":
        if settings.client_secret is None:
            raise ValueError("Client secret must be set for client_credentials flow")
        credentials = OAuthClientCredentials(
            token_url=settings.token_url,
            client_id=settings.client_id,
            client_secret=settings.client_secret,
            scopes=settings.scopes,
        )
    elif settings.login_flow == "interactive":
        credentials = OAuthDeviceCode(
            authority_url=settings.authority_uri, client_id=settings.client_id, scopes=settings.scopes
        )
    else:
        raise NotImplementedError(f"Unsupported login flow: {settings.login_flow!r}")

    return ClientConfig(
        client_name=settings.client_name, project=settings.project, base_url=settings.base_url, credentials=credentials
    )


def get_cognite_client() -> CogniteClient:
    settings = Settings()
    return CogniteClient(get_client_config(settings.cognite))



from __future__ import annotations

from collections.abc import Sequence
from typing import overload
import warnings

from cognite.client import CogniteClient
from cognite.client import data_modeling as dm

from cognite.powerops.client._generated.v1.data_classes._core import DEFAULT_INSTANCE_SPACE
from cognite.powerops.client._generated.v1.data_classes import (
    DomainModelCore,
    DomainModelWrite,
    ResourcesWriteResult,
    BidConfigurationWater,
    BidConfigurationWaterWrite,
    BidConfigurationWaterList,
    BidConfigurationWaterWriteList,
)
from cognite.powerops.client._generated.v1.data_classes._bid_configuration_water import (
    _create_bid_configuration_water_filter,
)
from ._core import (
    DEFAULT_LIMIT_READ,
    DEFAULT_QUERY_LIMIT,
    Aggregations,
    NodeAPI,
    SequenceNotStr,
    QueryStep,
    QueryBuilder,
)
from .bid_configuration_water_plants import BidConfigurationWaterPlantsAPI
from .bid_configuration_water_watercourses import BidConfigurationWaterWatercoursesAPI
from .bid_configuration_water_query import BidConfigurationWaterQueryAPI


class BidConfigurationWaterAPI(NodeAPI[BidConfigurationWater, BidConfigurationWaterWrite, BidConfigurationWaterList]):
    def __init__(self, client: CogniteClient, view_by_read_class: dict[type[DomainModelCore], dm.ViewId]):
        view_id = view_by_read_class[BidConfigurationWater]
        super().__init__(
            client=client,
            sources=view_id,
            class_type=BidConfigurationWater,
            class_list=BidConfigurationWaterList,
            class_write_list=BidConfigurationWaterWriteList,
            view_by_read_class=view_by_read_class,
        )
        self._view_id = view_id
        self.plants_edge = BidConfigurationWaterPlantsAPI(client)
        self.watercourses_edge = BidConfigurationWaterWatercoursesAPI(client)

    def __call__(
        self,
        market_configuration: str | tuple[str, str] | list[str] | list[tuple[str, str]] | None = None,
        method: str | tuple[str, str] | list[str] | list[tuple[str, str]] | None = None,
        price_area: str | tuple[str, str] | list[str] | list[tuple[str, str]] | None = None,
        external_id_prefix: str | None = None,
        space: str | list[str] | None = None,
        limit: int | None = DEFAULT_QUERY_LIMIT,
        filter: dm.Filter | None = None,
    ) -> BidConfigurationWaterQueryAPI[BidConfigurationWaterList]:
        """Query starting at bid configuration waters.

        Args:
            market_configuration: The market configuration to filter on.
            method: The method to filter on.
            price_area: The price area to filter on.
            external_id_prefix: The prefix of the external ID to filter on.
            space: The space to filter on.
            limit: Maximum number of bid configuration waters to return. Defaults to 25. Set to -1, float("inf") or None to return all items.
            filter: (Advanced) If the filtering available in the above is not sufficient, you can write your own filtering which will be ANDed with the filter above.

        Returns:
            A query API for bid configuration waters.

        """
        has_data = dm.filters.HasData(views=[self._view_id])
        filter_ = _create_bid_configuration_water_filter(
            self._view_id,
            market_configuration,
            method,
            price_area,
            external_id_prefix,
            space,
            (filter and dm.filters.And(filter, has_data)) or has_data,
        )
        builder = QueryBuilder(BidConfigurationWaterList)
        return BidConfigurationWaterQueryAPI(self._client, builder, self._view_by_read_class, filter_, limit)

    def apply(
        self,
        bid_configuration_water: BidConfigurationWaterWrite | Sequence[BidConfigurationWaterWrite],
        replace: bool = False,
        write_none: bool = False,
    ) -> ResourcesWriteResult:
        """Add or update (upsert) bid configuration waters.

        Note: This method iterates through all nodes and timeseries linked to bid_configuration_water and creates them including the edges
        between the nodes. For example, if any of `plants` or `watercourses` are set, then these
        nodes as well as any nodes linked to them, and all the edges linking these nodes will be created.

        Args:
            bid_configuration_water: Bid configuration water or sequence of bid configuration waters to upsert.
            replace (bool): How do we behave when a property value exists? Do we replace all matching and existing values with the supplied values (true)?
                Or should we merge in new values for properties together with the existing values (false)? Note: This setting applies for all nodes or edges specified in the ingestion call.
            write_none (bool): This method, will by default, skip properties that are set to None. However, if you want to set properties to None,
                you can set this parameter to True. Note this only applies to properties that are nullable.
        Returns:
            Created instance(s), i.e., nodes, edges, and time series.

        Examples:

            Create a new bid_configuration_water:

                >>> from cognite.powerops.client._generated.v1 import PowerOpsModelsV1Client
                >>> from cognite.powerops.client._generated.v1.data_classes import BidConfigurationWaterWrite
                >>> client = PowerOpsModelsV1Client()
                >>> bid_configuration_water = BidConfigurationWaterWrite(external_id="my_bid_configuration_water", ...)
                >>> result = client.bid_configuration_water.apply(bid_configuration_water)

        """
        warnings.warn(
            "The .apply method is deprecated and will be removed in v1.0. "
            "Please use the .upsert method on the client instead. This means instead of "
            "`my_client.bid_configuration_water.apply(my_items)` please use `my_client.upsert(my_items)`."
            "The motivation is that all apply methods are the same, and having one apply method per API "
            " class encourages users to create items in small batches, which is inefficient."
            "In addition, .upsert method is more descriptive of what the method does.",
            UserWarning,
            stacklevel=2,
        )
        return self._apply(bid_configuration_water, replace, write_none)

    def delete(
        self, external_id: str | SequenceNotStr[str], space: str = DEFAULT_INSTANCE_SPACE
    ) -> dm.InstancesDeleteResult:
        """Delete one or more bid configuration water.

        Args:
            external_id: External id of the bid configuration water to delete.
            space: The space where all the bid configuration water are located.

        Returns:
            The instance(s), i.e., nodes and edges which has been deleted. Empty list if nothing was deleted.

        Examples:

            Delete bid_configuration_water by id:

                >>> from cognite.powerops.client._generated.v1 import PowerOpsModelsV1Client
                >>> client = PowerOpsModelsV1Client()
                >>> client.bid_configuration_water.delete("my_bid_configuration_water")
        """
        warnings.warn(
            "The .delete method is deprecated and will be removed in v1.0. "
            "Please use the .delete method on the client instead. This means instead of "
            "`my_client.bid_configuration_water.delete(my_ids)` please use `my_client.delete(my_ids)`."
            "The motivation is that all delete methods are the same, and having one delete method per API "
            " class encourages users to delete items in small batches, which is inefficient.",
            UserWarning,
            stacklevel=2,
        )
        return self._delete(external_id, space)

    @overload
    def retrieve(self, external_id: str, space: str = DEFAULT_INSTANCE_SPACE) -> BidConfigurationWater | None: ...

    @overload
    def retrieve(
        self, external_id: SequenceNotStr[str], space: str = DEFAULT_INSTANCE_SPACE
    ) -> BidConfigurationWaterList: ...

    def retrieve(
        self, external_id: str | SequenceNotStr[str], space: str = DEFAULT_INSTANCE_SPACE
    ) -> BidConfigurationWater | BidConfigurationWaterList | None:
        """Retrieve one or more bid configuration waters by id(s).

        Args:
            external_id: External id or list of external ids of the bid configuration waters.
            space: The space where all the bid configuration waters are located.

        Returns:
            The requested bid configuration waters.

        Examples:

            Retrieve bid_configuration_water by id:

                >>> from cognite.powerops.client._generated.v1 import PowerOpsModelsV1Client
                >>> client = PowerOpsModelsV1Client()
                >>> bid_configuration_water = client.bid_configuration_water.retrieve("my_bid_configuration_water")

        """
        return self._retrieve(
            external_id,
            space,
            retrieve_edges=True,
            edge_api_name_type_direction_view_id_penta=[
                (
                    self.plants_edge,
                    "plants",
                    dm.DirectRelationReference("sp_powerops_types", "BidConfiguration.plants"),
                    "outwards",
                    dm.ViewId("sp_powerops_models", "Plant", "1"),
                ),
                (
                    self.watercourses_edge,
                    "watercourses",
                    dm.DirectRelationReference("sp_powerops_types", "BidConfiguration.watercourses"),
                    "outwards",
                    dm.ViewId("sp_powerops_models", "Watercourse", "1"),
                ),
            ],
        )

    def list(
        self,
        market_configuration: str | tuple[str, str] | list[str] | list[tuple[str, str]] | None = None,
        method: str | tuple[str, str] | list[str] | list[tuple[str, str]] | None = None,
        price_area: str | tuple[str, str] | list[str] | list[tuple[str, str]] | None = None,
        external_id_prefix: str | None = None,
        space: str | list[str] | None = None,
        limit: int | None = DEFAULT_LIMIT_READ,
        filter: dm.Filter | None = None,
        retrieve_edges: bool = True,
    ) -> BidConfigurationWaterList:
        """List/filter bid configuration waters

        Args:
            market_configuration: The market configuration to filter on.
            method: The method to filter on.
            price_area: The price area to filter on.
            external_id_prefix: The prefix of the external ID to filter on.
            space: The space to filter on.
            limit: Maximum number of bid configuration waters to return. Defaults to 25. Set to -1, float("inf") or None to return all items.
            filter: (Advanced) If the filtering available in the above is not sufficient, you can write your own filtering which will be ANDed with the filter above.
            retrieve_edges: Whether to retrieve `plants` or `watercourses` external ids for the bid configuration waters. Defaults to True.

        Returns:
            List of requested bid configuration waters

        Examples:

            List bid configuration waters and limit to 5:

                >>> from cognite.powerops.client._generated.v1 import PowerOpsModelsV1Client
                >>> client = PowerOpsModelsV1Client()
                >>> bid_configuration_waters = client.bid_configuration_water.list(limit=5)

        """
        filter_ = _create_bid_configuration_water_filter(
            self._view_id,
            market_configuration,
            method,
            price_area,
            external_id_prefix,
            space,
            filter,
        )

        return self._list(
            limit=limit,
            filter=filter_,
            retrieve_edges=retrieve_edges,
            edge_api_name_type_direction_view_id_penta=[
                (
                    self.plants_edge,
                    "plants",
                    dm.DirectRelationReference("sp_powerops_types", "BidConfiguration.plants"),
                    "outwards",
                    dm.ViewId("sp_powerops_models", "Plant", "1"),
                ),
                (
                    self.watercourses_edge,
                    "watercourses",
                    dm.DirectRelationReference("sp_powerops_types", "BidConfiguration.watercourses"),
                    "outwards",
                    dm.ViewId("sp_powerops_models", "Watercourse", "1"),
                ),
            ],
        )



from __future__ import annotations

from pathlib import Path

from cognite.client import ClientConfig, CogniteClient

from cognite.powerops.utils.cdf import Settings, get_client_config
from cognite.powerops.utils.serialization import read_toml_file

from ._generated.afrr_bid import AFRRBidAPI
from ._generated.assets import PowerAssetAPI
from ._generated.cogshop1 import CogShop1Client
from ._generated.day_ahead_bid import DayAheadBidAPI
from ._generated.v1 import PowerOpsModelsV1Client
from .data_set_api import DataSetsAPI
from .shop.api.dayahead_trigger_api import DayaheadTriggerAPI
from .shop.shop_run_api import SHOPRunAPI


class PowerOpsClient:
    def __init__(
        self,
        read_dataset: str,
        write_dataset: str,
        cogshop_version: str,
        config: ClientConfig,
        monitor_dataset: str | None = None,
    ):
        self.cdf = CogniteClient(config)
        self.cogshop_version = cogshop_version
        self.datasets = DataSetsAPI(self.cdf, read_dataset, write_dataset, monitor_dataset)
        self.cog_shop1 = CogShop1Client(self.cdf)
        self.assets = PowerAssetAPI(self.cdf)
        self.afrr_bid = AFRRBidAPI(self.cdf)
        self.day_ahead_bid = DayAheadBidAPI(self.cdf)
        self.shop = SHOPRunAPI(self.cdf, self.datasets.write_dataset_id, cogshop_version)
        self.workflow = DayaheadTriggerAPI(self.cdf, self.datasets.write_dataset_id, cogshop_version)
        self.v1 = PowerOpsModelsV1Client(self.cdf)

    def _apis(self) -> dict[str, str]:
        return {
            "cdf": "The regular Cognite Client",
            "cog_shop1": "The CogSHOP client, this is used by cogshop",
            "assets": "The PowerOps Assets model. For example, plants, generators etc",
            "afrr_bid": "The AFRR bid model, the model used to represent AFRR bids",
            "production": "(will be deprecated) The production model",
            "day_ahead_bid": "The day ahead bid model, the model used to represent day-ahead bids",
            "shop": "The shop model, this is used to trigger individual SHOP runs",
            "workflow": "The workflow model, this is used to trigger set of SHOP runs",
        }

    def _repr_html_(self) -> str:
        return (
            "<strong>PowerOpsClient:</strong><ul>"
            + "".join([f"<li><strong><em>.{k}</em></strong>: {v}</li>" for k, v in self._apis().items()])
            + "</ul>"
        )

    def __str__(self):
        return f"PowerOpsClient with {', '.join(map(lambda a: '.' + a, self._apis().keys()))} APIs"

    @classmethod
    def from_client(
        cls,
        client: CogniteClient,
        *,
        read_dataset: str | None = None,
        write_dataset: str | None = None,
        cogshop_version: str | None = None,
        monitor_dataset: str | None = None,
    ) -> PowerOpsClient:
        """
        Create a PowerOpsClient from a CogniteClient object.

        This uses default values for the read and write data sets, cogshop version and monitor data set.

        Args:
            client: The CogniteClient object.
            read_dataset: externalId of read data set. Optional, by default loaded from the settings object.
            write_dataset: externalId of write data set. Optional, by default loaded from the settings object.
            monitor_dataset: externalId of monitor data set. Optional, by default loaded from the settings object.
            cogshop_version: tag for the "cog-shop" Docker image. Optional, by default loaded from the settings object.

        Returns:
            A PowerOpsClient object.
        """
        settings = Settings()
        return PowerOpsClient(
            config=client.config,
            read_dataset=read_dataset if read_dataset is not None else settings.powerops.read_dataset,
            write_dataset=write_dataset if write_dataset is not None else settings.powerops.write_dataset,
            monitor_dataset=monitor_dataset if monitor_dataset is not None else settings.powerops.monitor_dataset,
            cogshop_version=cogshop_version if cogshop_version is not None else settings.powerops.cogshop_version,
        )

    @classmethod
    def from_settings(
        cls,
        settings: Settings | None = None,
        *,
        config: ClientConfig | None = None,
        read_dataset: str | None = None,
        write_dataset: str | None = None,
        cogshop_version: str | None = None,
        monitor_dataset: str | None = None,
    ) -> PowerOpsClient:
        """
        Create a PowerOpsClient from a Settings object.


        Args:
            settings: The settings object. If no settings object is provided, the settings object is loaded from
                the environment. When loading the Settings object from the environment,
                the environment variable `SETTINGS_FILES` is used to
                specify which files to load. The default value is `settings.toml;.secrets.toml`.
            config: The client config object. Optional, by default it is loaded from the settings object.
            read_dataset: externalId of read data set. Optional, by default loaded from the settings object.
            write_dataset: externalId of write data set. Optional, by default loaded from the settings object.
            monitor_dataset: externalId of monitor data set. Optional, by default loaded from the settings object.
            cogshop_version: tag for the "cog-shop" Docker image. Optional, by default loaded from the settings object.

        Returns:
            A PowerOpsClient object.
        """
        settings = settings or Settings()

        client_config = config if config is not None else get_client_config(settings.cognite)

        return PowerOpsClient(
            config=client_config,
            read_dataset=read_dataset if read_dataset is not None else settings.powerops.read_dataset,
            write_dataset=write_dataset if write_dataset is not None else settings.powerops.write_dataset,
            monitor_dataset=monitor_dataset if monitor_dataset is not None else settings.powerops.monitor_dataset,
            cogshop_version=cogshop_version if cogshop_version is not None else settings.powerops.cogshop_version,
        )

    @classmethod
    def from_toml(cls, toml_file: str | Path) -> PowerOpsClient:
        """
        Create a PowerOpsClient from a TOML file.

        Args:
            toml_file: Path to the TOML file.

        Returns:
            A PowerOpsClient object.
        """
        content = read_toml_file(toml_file)
        return cls.from_settings(Settings.model_validate(content))



# Mypy does not understand the pydantic classes that allows both alias and name to be used in population
# https://github.com/pydantic/pydantic/issues/3923
# mypy: disable-error-code="call-arg"
from __future__ import annotations

from math import floor, log10
from pathlib import Path
from typing import Any, Optional

from cognite.powerops.client._generated.v1.data_classes import (
    GeneratorEfficiencyCurveWrite,
    GeneratorWrite,
    PlantWrite,
    PriceAreaWrite,
    ReservoirWrite,
    TurbineEfficiencyCurveWrite,
    WatercourseWrite,
)
from cognite.powerops.utils.serialization import load_yaml

__all__ = ["PowerAssetImporter"]


class PowerAssetImporter:
    p_min_fallback = 0.0
    p_max_fallback = 10_000_000_000_000_000_000.0
    head_loss_factor_fallback = 0.0
    connection_losses_fallback = 0.0
    inlet_reservoir_fallback = ""
    default_shop_penalty_limit = 42000
    default_timezone = "Europe/Oslo"

    def __init__(
        self,
        shop_model_by_directory: dict[str, Any],
        generator_times_series_mappings: Optional[list[dict[str, Any]]] = None,
        plant_time_series_mappings: Optional[list[dict[str, Any]]] = None,
        watercourses: Optional[list[dict[str, Any]]] = None,
    ):
        self.shop_model_by_directory = shop_model_by_directory
        self.times_series_by_generator_name = {
            entry["generator_name"]: {k: str(v) for k, v in entry.items()}
            for entry in generator_times_series_mappings or []
        }
        self.times_series_by_plant_name = {
            entry["plant_name"]: {k: str(v) for k, v in entry.items()} for entry in plant_time_series_mappings or []
        }
        self.watercourse_by_directory = {entry["directory"]: entry for entry in watercourses or []}

    @classmethod
    def from_directory(cls, directory: Path, file_name: str = "model_raw") -> PowerAssetImporter:
        shop_model_files = list(directory.glob(f"**/{file_name}.yaml"))
        shop_model_by_watercourse = {
            file.parent.name: load_yaml(file, expected_return_type="dict") for file in shop_model_files
        }

        generator_mapping_file = directory / "generator_time_series_mappings.yaml"
        generator_mappings = (
            load_yaml(generator_mapping_file, expected_return_type="list") if generator_mapping_file.exists() else None
        )

        plant_mapping_file = directory / "plant_time_series_mappings.yaml"
        plant_mappings = (
            load_yaml(plant_mapping_file, expected_return_type="list") if plant_mapping_file.exists() else None
        )

        watercourses_file = directory / "watercourses.yaml"
        watercourses = load_yaml(watercourses_file, expected_return_type="list") if watercourses_file.exists() else None

        return cls(shop_model_by_watercourse, generator_mappings, plant_mappings, watercourses)

    def to_power_assets(self) -> list[GeneratorWrite | ReservoirWrite | PlantWrite | WatercourseWrite | PriceAreaWrite]:
        assets_by_xid: dict[str, GeneratorWrite | ReservoirWrite | PlantWrite | WatercourseWrite | PriceAreaWrite] = {}
        for watercourse_dir, shop_model in self.shop_model_by_directory.items():
            watercourse_assets = self._shop_model_to_watercourse_assets(shop_model, watercourse_dir, assets_by_xid)
            assets_by_xid.update(watercourse_assets)

        return list(assets_by_xid.values())

    def _shop_model_to_watercourse_assets(
        self,
        shop_model: dict[str, Any],
        watercourse_dir: str,
        existing: dict[str, GeneratorWrite | ReservoirWrite | PlantWrite | WatercourseWrite | PriceAreaWrite],
    ) -> dict[str, GeneratorWrite | ReservoirWrite | PlantWrite | WatercourseWrite | PriceAreaWrite]:
        assets_by_xid: dict[str, GeneratorWrite | ReservoirWrite | PlantWrite | WatercourseWrite | PriceAreaWrite] = {}
        try:
            watercourse_config = self.watercourse_by_directory[watercourse_dir]
        except KeyError as e:
            raise ValueError(
                f"Watercourse directory {watercourse_dir} does not exist in the watercourses.yaml configuration"
            ) from e
        watercourse_external_id = f"watercourse_{watercourse_config['name']}"
        plant_display_name_and_order = watercourse_config.get("plant_display_names_and_order", {})
        reservoir_display_name_and_order = watercourse_config.get("reservoir_display_names_and_order", {})

        generator_by_name = {}
        for name, data in shop_model["model"]["generator"].items():
            generator = self._to_generator(name, data)
            if generator.external_id in existing:
                raise ValueError(f"Generator with external id {generator.external_id} already exists")
            generator_by_name[generator.name] = generator
            assets_by_xid[generator.external_id] = generator

        reservoir_by_name = {}
        for name in shop_model["model"]["reservoir"].keys():
            reservoir = self._to_reservoir(name, reservoir_display_name_and_order)
            if reservoir.external_id in existing:
                raise ValueError(f"Reservoir with external id {reservoir.external_id} already exists")
            reservoir_by_name[reservoir.name] = reservoir
            assets_by_xid[reservoir.external_id] = reservoir

        plants = []
        for name, data in shop_model["model"]["plant"].items():
            plant = self._to_plant(
                name,
                data,
                plant_display_name_and_order,
                watercourse_external_id,
                reservoir_by_name,
                generator_by_name,
                shop_model,
            )
            if plant.external_id in existing:
                raise ValueError(f"Plant with external id {plant.external_id} already exists")
            plants.append(plant)
            assets_by_xid[plant.external_id] = plant

        data = watercourse_config
        watercourse = WatercourseWrite(
            external_id=watercourse_external_id,
            name=data["name"],
            production_obligation=data.get("production_obligation_ts_ext_ids", []),
            penalty_limit=data.get("shop_penalty_limit", self.default_shop_penalty_limit),
            plants=plants,
        )
        if watercourse.external_id in existing:
            raise ValueError(f"Watercourse with external id {watercourse.external_id} already exists")
        assets_by_xid[watercourse.external_id] = watercourse

        for price_area in watercourse_config.get("market_to_price_area", {}).values():
            if price_area in existing:
                raise ValueError(f"Price area with external id {price_area} already exists")
            price_area = PriceAreaWrite(
                external_id=f"price_area_{price_area}", name=price_area, timezone=self.default_timezone
            )
            assets_by_xid[price_area.external_id] = price_area

        return assets_by_xid

    def _to_generator(self, name: str, data: dict) -> GeneratorWrite:
        curve = data["gen_eff_curve"]
        efficiency_curve = GeneratorEfficiencyCurveWrite(
            external_id=f"{name}_efficiency_curve", ref=curve["ref"], power=curve["x"], efficiency=curve["y"]
        )
        turbine_curves = []
        for curve in data["turb_eff_curves"]:
            turbine_curve = TurbineEfficiencyCurveWrite(
                external_id=f"{name}_turbine_eff_curve_ref_{curve['ref']}",
                head=curve["ref"],
                flow=curve["x"],
                efficiency=curve["y"],
            )
            turbine_curves.append(turbine_curve)

        # Start cost is assumed to be a dictionary with (timestamp, value) key-value pairs
        # We assume it is constant and thus use only the first value.
        startcost = next(iter(data["startcost"].values()))

        generator_timeseries = self.times_series_by_generator_name.get(name, {})

        return GeneratorWrite(
            external_id=f"generator_{name}",
            name=name,
            p_min=data["p_min"],
            penstock=data["penstock"],
            start_cost=startcost,
            start_stop_cost=generator_timeseries.get("start_stop_cost"),
            is_available_time_series=generator_timeseries.get("is_available"),
            efficiency_curve=efficiency_curve,
            turbine_curves=turbine_curves,
        )

    def _to_plant(
        self,
        name: str,
        data: dict,
        plant_display_name_and_order: dict,
        watercourse_xid: str,
        reservoir_by_name: dict[str, ReservoirWrite],
        generator_by_name: dict[str, GeneratorWrite],
        shop_model: dict,
    ) -> PlantWrite:
        plant_timeseries = self.times_series_by_plant_name.get(name, {})
        display_name, order = plant_display_name_and_order.get(name, (name, 999))

        all_connections = shop_model["connections"]
        all_junctions = shop_model["model"].get("junction", {})
        all_tunnels = shop_model["model"].get("tunnel", {})
        inlet_reservoir_name, connection_losses = self._plant_to_inlet_reservoir_with_losses(
            name, all_connections, all_junctions, all_tunnels, set(reservoir_by_name.keys())
        )

        plant_generators: dict[str, GeneratorWrite] = {}
        for connection in all_connections:
            if (
                connection.get("from_type") == "plant"
                and connection["from"] == name
                and (gen := generator_by_name.get(connection["to"]))
            ):
                plant_generators[gen.name] = gen
            elif (
                connection.get("to_type") == "plant"
                and connection["to"] == name
                and (gen := generator_by_name.get(connection["from"]))
            ):
                plant_generators[gen.name] = gen

        return PlantWrite(
            external_id=f"plant_{name}",
            name=name,
            display_name=display_name,
            outlet_level=float(data.get("outlet_line", 0)),
            p_min=float(data.get("p_min", self.p_min_fallback)),
            p_max=float(data.get("p_max", self.p_max_fallback)),
            ordering=order,
            penstock_head_loss_factors={
                str(penstock_number): float(loss_factor)
                for penstock_number, loss_factor in enumerate(
                    data.get("penstock_loss", [self.head_loss_factor_fallback]), start=1
                )
            },
            head_loss_factor=float(data.get("main_loss", [self.head_loss_factor_fallback])[0]),
            connection_losses=connection_losses,
            water_value_time_series=plant_timeseries.get("water_value"),
            inlet_level_time_series=plant_timeseries.get("inlet_reservoir_level"),
            outlet_level_time_series=plant_timeseries.get("outlet_reservoir_level"),
            p_min_time_series=plant_timeseries.get("p_min"),
            p_max_time_series=plant_timeseries.get("p_max"),
            feeding_fee_time_series=plant_timeseries.get("feeding_fee"),
            head_direct_time_series=plant_timeseries.get("head_direct"),
            watercourse=watercourse_xid,
            inlet_reservoir=reservoir_by_name.get(inlet_reservoir_name),
            generators=list(plant_generators.values()),
        )

    @classmethod
    def _to_reservoir(cls, name: str, reservoir_display_name_and_order: dict) -> ReservoirWrite:
        display_name, order = reservoir_display_name_and_order.get(name, (name, 999))

        return ReservoirWrite(external_id=f"reservoir_{name}", name=name, display_name=display_name, ordering=order)

    # Todo - Refactor/Test coverage of this function:
    #  https://cognitedata.atlassian.net/browse/POWEROPS-2224?atlOrigin=eyJpIjoiNGFhYWQxMGU0NTE3NGEzNDlhZTBkN2Y5NDhkYTczYmYiLCJwIjoiaiJ9
    def _plant_to_inlet_reservoir_with_losses(
        self, plant_name: str, all_connections: list[dict], all_junctions: dict, all_tunnels: dict, reservoirs: set[str]
    ) -> tuple[str, float]:
        """Search for a reservoir connected to a plant, starting from the plant and searching breadth first.

        Parameters
        ----------
        plant_name : str
            The plant we want to find a connection from
        all_connections : list[dict]
            All connections in the model.
        reservoirs : dict
            All reservoirs in the model. Keys are reservoir names.

        Returns
        -------
        tuple[str, float]
            The name of the reservoir connected to the plant, or None if no reservoir was found.
        """

        def get_connection_path_from_last_visited(visited_paths: list, last_visited_id: int) -> list | None:
            """Return the correct sequence of visited connections based on the last visited connection among
            the list of connection paths visited

                Parameters
                ----------
                visited_paths : list
                     A list that holds the lists of each visited connection path. A connection path will be a list of
                     connection IDs, e.g. [1,2,3] means connection with ID 1 was first visited, then 2, then 3
                last_visited_id: int
                    The ID of the connection that was last visited. This will be the last item in one of the lists of
                    visited paths

                Returns
                -------
                list
                    The path or sequence of connections that has the last_visited_id as its
                    last visited connection among the visited_paths
            """
            for connection in visited_paths:
                if connection[-1] == last_visited_id:
                    connection_path_index = visited_paths.index(connection)
                    return visited_paths[connection_path_index]
            return None

        def calculate_losses_from_connection_path(
            all_junctions: dict, all_tunnels: dict, connection_by_id: dict, connection_path: list[int]
        ):
            """Loop through connections in the connection path, retrieve the losses for that connection among the
            all_junctions or all_tunnels based on the type of connection, and sum up the total losses from
            the connection path
            """
            sum_losses = 0
            order_to_loss_factor_key = {0: "loss_factor_1", 1: "loss_factor_2"}
            for connection_id in connection_path:
                connection = connection_by_id[connection_id]
                connection_name = connection["to"]
                if connection_name in all_junctions:
                    if connection.get("order") in order_to_loss_factor_key:
                        junction_losses = all_junctions[connection_name]
                        loss_order = connection["order"]
                        sum_losses += junction_losses[order_to_loss_factor_key[loss_order]]
                elif connection_name in all_tunnels:
                    tunnel_loss = all_tunnels[connection_name].get("loss_factor", 0)
                    sum_losses += tunnel_loss
            return sum_losses

        queue = []
        connection_by_id = dict(enumerate(all_connections))
        track_connection_paths = []
        last_connection_id = None

        for connection_id, connection in connection_by_id.items():
            if (
                connection["to"] == plant_name and connection.get("to_type", "plant") == "plant"
            ):  # if to_type is specified, it must be "plant"
                queue.append((connection_id, connection))
                track_connection_paths.append([connection_id])  # add the first connection to the path
                break
        visited = []
        while queue:
            connection_id, connection = queue.pop(0)
            if connection not in visited:
                # Check if the given connection is from a reservoir
                # If we have "from_type" we can check directly if the object is a reservoir
                try:
                    if connection["from_type"] == "reservoir":
                        inlet_reservoir = connection["from"]
                        last_connection_id = connection_id
                        break
                # If we don't have "from_type" we have to check if the name of the object is in the
                # list of reservoirs
                except KeyError:
                    if connection["from"] in reservoirs:
                        inlet_reservoir = connection["from"]
                        last_connection_id = connection_id
                        break

                visited.append(connection)
                for candidate_connection_id, candidate_connection in connection_by_id.items():
                    # if the candidate connection is extension from the current connection, traverse it
                    if candidate_connection["to"] == connection["from"]:
                        queue.append((candidate_connection_id, candidate_connection))
                        new_path_list = get_connection_path_from_last_visited(track_connection_paths, connection_id)
                        if new_path_list:
                            track_connection_paths.append([*new_path_list, candidate_connection_id])

        if last_connection_id is None:
            return self.inlet_reservoir_fallback, self.connection_losses_fallback  # TODO: raise an error here instead?

        connection_path = get_connection_path_from_last_visited(track_connection_paths, last_connection_id) or []

        connection_losses = calculate_losses_from_connection_path(
            all_junctions, all_tunnels, connection_by_id, connection_path
        )

        return (inlet_reservoir, self._round_sig(connection_losses, 4) if connection_losses else connection_losses)

    @staticmethod
    def _round_sig(x: float, sig: int = 2):
        return round(x, sig - int(floor(log10(abs(x)))) - 1)



from __future__ import annotations

import warnings
from typing import TYPE_CHECKING, Any, Literal, Optional, Union

from cognite.client import data_modeling as dm
from pydantic import Field

from ._core import (
    DEFAULT_INSTANCE_SPACE,
    DataRecordWrite,
    DomainModel,
    DomainModelCore,
    DomainModelWrite,
    DomainModelWriteList,
    DomainModelList,
    DomainRelationWrite,
    ResourcesWrite,
)

if TYPE_CHECKING:
    from ._mapping import Mapping, MappingWrite
    from ._watercourse_shop import WatercourseShop, WatercourseShopWrite


__all__ = [
    "ModelTemplate",
    "ModelTemplateWrite",
    "ModelTemplateApply",
    "ModelTemplateList",
    "ModelTemplateWriteList",
    "ModelTemplateApplyList",
    "ModelTemplateFields",
    "ModelTemplateTextFields",
]


ModelTemplateTextFields = Literal["version_", "shop_version", "model", "extra_files"]
ModelTemplateFields = Literal["version_", "shop_version", "model", "cog_shop_files_config", "extra_files"]

_MODELTEMPLATE_PROPERTIES_BY_FIELD = {
    "version_": "version",
    "shop_version": "shopVersion",
    "model": "model",
    "cog_shop_files_config": "cogShopFilesConfig",
    "extra_files": "extraFiles",
}


class ModelTemplate(DomainModel, protected_namespaces=()):
    """This represents the reading version of model template.

    It is used to when data is retrieved from CDF.

    Args:
        space: The space where the node is located.
        external_id: The external id of the model template.
        data_record: The data record of the model template node.
        version_: The version of the model
        shop_version: The version of SHOP to run
        watercourse: The watercourse to run the model for
        model: The shop model file to use as template before applying base mapping
        cog_shop_files_config: Configuration for in what order to load the various files into pyshop
        extra_files: Extra files related to a model template
        base_mappings: The base mappings for the model
    """

    space: str = DEFAULT_INSTANCE_SPACE
    node_type: Union[dm.DirectRelationReference, None] = dm.DirectRelationReference(
        "sp_powerops_types", "ModelTemplate"
    )
    version_: Optional[str] = Field(None, alias="version")
    shop_version: str = Field(alias="shopVersion")
    watercourse: Union[WatercourseShop, str, dm.NodeId, None] = Field(None, repr=False)
    model: Union[str, None] = None
    cog_shop_files_config: Optional[list[dict]] = Field(None, alias="cogShopFilesConfig")
    extra_files: Optional[list[str]] = Field(None, alias="extraFiles")
    base_mappings: Union[list[Mapping], list[str], None] = Field(default=None, repr=False, alias="baseMappings")

    def as_write(self) -> ModelTemplateWrite:
        """Convert this read version of model template to the writing version."""
        return ModelTemplateWrite(
            space=self.space,
            external_id=self.external_id,
            data_record=DataRecordWrite(existing_version=self.data_record.version),
            version_=self.version_,
            shop_version=self.shop_version,
            watercourse=self.watercourse.as_write() if isinstance(self.watercourse, DomainModel) else self.watercourse,
            model=self.model,
            cog_shop_files_config=self.cog_shop_files_config,
            extra_files=self.extra_files,
            base_mappings=[
                base_mapping.as_write() if isinstance(base_mapping, DomainModel) else base_mapping
                for base_mapping in self.base_mappings or []
            ],
        )

    def as_apply(self) -> ModelTemplateWrite:
        """Convert this read version of model template to the writing version."""
        warnings.warn(
            "as_apply is deprecated and will be removed in v1.0. Use as_write instead.",
            UserWarning,
            stacklevel=2,
        )
        return self.as_write()


class ModelTemplateWrite(DomainModelWrite, protected_namespaces=()):
    """This represents the writing version of model template.

    It is used to when data is sent to CDF.

    Args:
        space: The space where the node is located.
        external_id: The external id of the model template.
        data_record: The data record of the model template node.
        version_: The version of the model
        shop_version: The version of SHOP to run
        watercourse: The watercourse to run the model for
        model: The shop model file to use as template before applying base mapping
        cog_shop_files_config: Configuration for in what order to load the various files into pyshop
        extra_files: Extra files related to a model template
        base_mappings: The base mappings for the model
    """

    space: str = DEFAULT_INSTANCE_SPACE
    node_type: Union[dm.DirectRelationReference, None] = dm.DirectRelationReference(
        "sp_powerops_types", "ModelTemplate"
    )
    version_: Optional[str] = Field(None, alias="version")
    shop_version: str = Field(alias="shopVersion")
    watercourse: Union[WatercourseShopWrite, str, dm.NodeId, None] = Field(None, repr=False)
    model: Union[str, None] = None
    cog_shop_files_config: Optional[list[dict]] = Field(None, alias="cogShopFilesConfig")
    extra_files: Optional[list[str]] = Field(None, alias="extraFiles")
    base_mappings: Union[list[MappingWrite], list[str], None] = Field(default=None, repr=False, alias="baseMappings")

    def _to_instances_write(
        self,
        cache: set[tuple[str, str]],
        view_by_read_class: dict[type[DomainModelCore], dm.ViewId] | None,
        write_none: bool = False,
    ) -> ResourcesWrite:
        resources = ResourcesWrite()
        if self.as_tuple_id() in cache:
            return resources

        write_view = (view_by_read_class or {}).get(
            ModelTemplate, dm.ViewId("sp_powerops_models", "ModelTemplate", "1")
        )

        properties: dict[str, Any] = {}

        if self.version_ is not None or write_none:
            properties["version"] = self.version_

        if self.shop_version is not None:
            properties["shopVersion"] = self.shop_version

        if self.watercourse is not None:
            properties["watercourse"] = {
                "space": self.space if isinstance(self.watercourse, str) else self.watercourse.space,
                "externalId": self.watercourse if isinstance(self.watercourse, str) else self.watercourse.external_id,
            }

        if self.model is not None:
            properties["model"] = self.model

        if self.cog_shop_files_config is not None or write_none:
            properties["cogShopFilesConfig"] = self.cog_shop_files_config

        if self.extra_files is not None or write_none:
            properties["extraFiles"] = self.extra_files

        if properties:
            this_node = dm.NodeApply(
                space=self.space,
                external_id=self.external_id,
                existing_version=self.data_record.existing_version,
                type=self.node_type,
                sources=[
                    dm.NodeOrEdgeData(
                        source=write_view,
                        properties=properties,
                    )
                ],
            )
            resources.nodes.append(this_node)
            cache.add(self.as_tuple_id())

        edge_type = dm.DirectRelationReference("sp_powerops_types", "ModelTemplate.baseMappings")
        for base_mapping in self.base_mappings or []:
            other_resources = DomainRelationWrite.from_edge_to_resources(
                cache,
                start_node=self,
                end_node=base_mapping,
                edge_type=edge_type,
                view_by_read_class=view_by_read_class,
            )
            resources.extend(other_resources)

        if isinstance(self.watercourse, DomainModelWrite):
            other_resources = self.watercourse._to_instances_write(cache, view_by_read_class)
            resources.extend(other_resources)

        return resources


class ModelTemplateApply(ModelTemplateWrite):
    def __new__(cls, *args, **kwargs) -> ModelTemplateApply:
        warnings.warn(
            "ModelTemplateApply is deprecated and will be removed in v1.0. Use ModelTemplateWrite instead."
            "The motivation for this change is that Write is a more descriptive name for the writing version of the"
            "ModelTemplate.",
            UserWarning,
            stacklevel=2,
        )
        return super().__new__(cls)


class ModelTemplateList(DomainModelList[ModelTemplate]):
    """List of model templates in the read version."""

    _INSTANCE = ModelTemplate

    def as_write(self) -> ModelTemplateWriteList:
        """Convert these read versions of model template to the writing versions."""
        return ModelTemplateWriteList([node.as_write() for node in self.data])

    def as_apply(self) -> ModelTemplateWriteList:
        """Convert these read versions of primitive nullable to the writing versions."""
        warnings.warn(
            "as_apply is deprecated and will be removed in v1.0. Use as_write instead.",
            UserWarning,
            stacklevel=2,
        )
        return self.as_write()


class ModelTemplateWriteList(DomainModelWriteList[ModelTemplateWrite]):
    """List of model templates in the writing version."""

    _INSTANCE = ModelTemplateWrite


class ModelTemplateApplyList(ModelTemplateWriteList): ...


def _create_model_template_filter(
    view_id: dm.ViewId,
    version_: str | list[str] | None = None,
    version_prefix: str | None = None,
    shop_version: str | list[str] | None = None,
    shop_version_prefix: str | None = None,
    watercourse: str | tuple[str, str] | list[str] | list[tuple[str, str]] | None = None,
    external_id_prefix: str | None = None,
    space: str | list[str] | None = None,
    filter: dm.Filter | None = None,
) -> dm.Filter | None:
    filters = []
    if isinstance(version_, str):
        filters.append(dm.filters.Equals(view_id.as_property_ref("version"), value=version_))
    if version_ and isinstance(version_, list):
        filters.append(dm.filters.In(view_id.as_property_ref("version"), values=version_))
    if version_prefix is not None:
        filters.append(dm.filters.Prefix(view_id.as_property_ref("version"), value=version_prefix))
    if isinstance(shop_version, str):
        filters.append(dm.filters.Equals(view_id.as_property_ref("shopVersion"), value=shop_version))
    if shop_version and isinstance(shop_version, list):
        filters.append(dm.filters.In(view_id.as_property_ref("shopVersion"), values=shop_version))
    if shop_version_prefix is not None:
        filters.append(dm.filters.Prefix(view_id.as_property_ref("shopVersion"), value=shop_version_prefix))
    if watercourse and isinstance(watercourse, str):
        filters.append(
            dm.filters.Equals(
                view_id.as_property_ref("watercourse"),
                value={"space": DEFAULT_INSTANCE_SPACE, "externalId": watercourse},
            )
        )
    if watercourse and isinstance(watercourse, tuple):
        filters.append(
            dm.filters.Equals(
                view_id.as_property_ref("watercourse"), value={"space": watercourse[0], "externalId": watercourse[1]}
            )
        )
    if watercourse and isinstance(watercourse, list) and isinstance(watercourse[0], str):
        filters.append(
            dm.filters.In(
                view_id.as_property_ref("watercourse"),
                values=[{"space": DEFAULT_INSTANCE_SPACE, "externalId": item} for item in watercourse],
            )
        )
    if watercourse and isinstance(watercourse, list) and isinstance(watercourse[0], tuple):
        filters.append(
            dm.filters.In(
                view_id.as_property_ref("watercourse"),
                values=[{"space": item[0], "externalId": item[1]} for item in watercourse],
            )
        )
    if external_id_prefix is not None:
        filters.append(dm.filters.Prefix(["node", "externalId"], value=external_id_prefix))
    if isinstance(space, str):
        filters.append(dm.filters.Equals(["node", "space"], value=space))
    if space and isinstance(space, list):
        filters.append(dm.filters.In(["node", "space"], values=space))
    if filter:
        filters.append(filter)
    return dm.filters.And(*filters) if filters else None



from __future__ import annotations

import datetime
import warnings
from abc import abstractmethod
from collections import UserList
from collections.abc import Collection, Mapping
from dataclasses import dataclass, field
from typing import (
    Annotated,
    Callable,
    ClassVar,
    Generic,
    Optional,
    Any,
    Iterator,
    TypeVar,
    overload,
    Union,
)

import pandas as pd
from cognite.client import data_modeling as dm
from cognite.client.data_classes import TimeSeries as CogniteTimeSeries
from cognite.client.data_classes import TimeSeriesList
from cognite.client.data_classes.data_modeling.instances import (
    Instance,
    InstanceApply,
    InstanceCore,
    Properties,
    PropertyValue,
)
from pydantic import BaseModel, BeforeValidator, Extra, Field, model_validator
from pydantic.functional_serializers import PlainSerializer

TimeSeries = Annotated[
    CogniteTimeSeries,
    PlainSerializer(
        lambda v: v.dump(camel_case=True) if isinstance(v, CogniteTimeSeries) else v,
        return_type=dict,
        when_used="unless-none",
    ),
    BeforeValidator(lambda v: CogniteTimeSeries.load(v) if isinstance(v, dict) else v),
]


DEFAULT_INSTANCE_SPACE = "sp_powerops_instance"


@dataclass
class ResourcesWrite:
    nodes: dm.NodeApplyList = field(default_factory=lambda: dm.NodeApplyList([]))
    edges: dm.EdgeApplyList = field(default_factory=lambda: dm.EdgeApplyList([]))
    time_series: TimeSeriesList = field(default_factory=lambda: TimeSeriesList([]))

    def extend(self, other: ResourcesWrite) -> None:
        self.nodes.extend(other.nodes)
        self.edges.extend(other.edges)
        self.time_series.extend(other.time_series)


@dataclass
class ResourcesWriteResult:
    nodes: dm.NodeApplyResultList
    edges: dm.EdgeApplyResultList
    time_series: TimeSeriesList


# Arbitrary types are allowed to be able to use the TimeSeries class
class Core(BaseModel, arbitrary_types_allowed=True, populate_by_name=True):
    def to_pandas(self) -> pd.Series:
        return pd.Series(self.model_dump())

    def _repr_html_(self) -> str:
        """Returns HTML representation of DomainModel."""
        return self.to_pandas().to_frame("value")._repr_html_()  # type: ignore[operator]


class DomainModelCore(Core):
    space: str
    external_id: str = Field(min_length=1, max_length=255, alias="externalId")

    def as_tuple_id(self) -> tuple[str, str]:
        return self.space, self.external_id

    def as_direct_reference(self) -> dm.DirectRelationReference:
        return dm.DirectRelationReference(space=self.space, external_id=self.external_id)

    @classmethod
    @abstractmethod
    def from_instance(cls: type[T_DomainModelCore], instance: InstanceCore) -> T_DomainModelCore:
        raise NotImplementedError()


T_DomainModelCore = TypeVar("T_DomainModelCore", bound=DomainModelCore)


class DataRecord(BaseModel):
    """The data record represents the metadata of a node.

    Args:
        created_time: The created time of the node.
        last_updated_time: The last updated time of the node.
        deleted_time: If present, the deleted time of the node.
        version: The version of the node.
    """

    version: int
    last_updated_time: datetime.datetime
    created_time: datetime.datetime
    deleted_time: Optional[datetime.datetime] = None


class DomainModel(DomainModelCore):
    data_record: DataRecord

    def as_id(self) -> dm.NodeId:
        return dm.NodeId(space=self.space, external_id=self.external_id)

    @classmethod
    def from_instance(cls: type[T_DomainModel], instance: Instance) -> T_DomainModel:
        data = instance.dump(camel_case=False)
        node_type = data.pop("type", None)
        space = data.pop("space")
        external_id = data.pop("external_id")
        return cls(
            space=space,
            external_id=external_id,
            data_record=DataRecord(**data),
            node_type=node_type,
            **unpack_properties(instance.properties),
        )


T_DomainModel = TypeVar("T_DomainModel", bound=DomainModel)


class DataRecordWrite(BaseModel):
    """The data record represents the metadata of a node.

    Args:
        existing_version: Fail the ingestion request if the node version is greater than or equal to this value.
            If no existingVersion is specified, the ingestion will always overwrite any existing data for the edge (for the specified container or instance).
            If existingVersion is set to 0, the upsert will behave as an insert, so it will fail the bulk if the item already exists.
            If skipOnVersionConflict is set on the ingestion request, then the item will be skipped instead of failing the ingestion request.
    """

    existing_version: Optional[int] = None


T_DataRecord = TypeVar("T_DataRecord", bound=Union[DataRecord, DataRecordWrite])


class _DataRecordListCore(UserList, Generic[T_DataRecord]):
    def __init__(self, nodes: Collection[T_DataRecord] | None = None):
        super().__init__(nodes or [])

    # The dunder implementations are to get proper type hints
    def __iter__(self) -> Iterator[T_DataRecord]:
        return super().__iter__()

    @overload
    def __getitem__(self, item: int) -> T_DataRecord: ...

    @overload
    def __getitem__(
        self: type[_DataRecordListCore[T_DataRecord]], item: slice
    ) -> type[_DataRecordListCore[T_DataRecord]]: ...

    def __getitem__(self, item: int | slice) -> T_DataRecord | type[_DataRecordListCore[T_DataRecord]]:
        if isinstance(item, slice):
            return self.__class__(self.data[item])
        elif isinstance(item, int):
            return self.data[item]
        else:
            raise TypeError(f"Expected int or slice, got {type(item)}")

    def to_pandas(self) -> pd.DataFrame:
        """
        Convert the list of nodes to a pandas.DataFrame.

        Returns:
            A pandas.DataFrame with the nodes as rows.
        """
        df = pd.DataFrame([item.model_dump() for item in self])
        if df.empty:
            df = pd.DataFrame(columns=self._INSTANCE.model_fields)
        return df

    def _repr_html_(self) -> str:
        return self.to_pandas()._repr_html_()  # type: ignore[operator]


class DataRecordList(_DataRecordListCore[DataRecord]):
    _INSTANCE = DataRecord


class DataRecordWriteList(_DataRecordListCore[DataRecordWrite]):
    _INSTANCE = DataRecordWrite


class DomainModelWrite(DomainModelCore, extra=Extra.forbid, populate_by_name=True):
    external_id_factory: ClassVar[Optional[Callable[[type[DomainModelWrite], dict], str]]] = None
    data_record: DataRecordWrite = Field(default_factory=DataRecordWrite)

    def to_instances_write(
        self, view_by_read_class: dict[type[DomainModelCore], dm.ViewId] | None = None, write_none: bool = False
    ) -> ResourcesWrite:
        return self._to_instances_write(set(), view_by_read_class, write_none)

    def to_instances_apply(
        self, view_by_read_class: dict[type[DomainModelCore], dm.ViewId] | None = None, write_none: bool = False
    ) -> ResourcesWrite:
        warnings.warn(
            "to_instances_apply is deprecated and will be removed in v1.0. Use to_instances_write instead.",
            UserWarning,
            stacklevel=2,
        )
        return self.to_instances_write(view_by_read_class, write_none)

    @abstractmethod
    def _to_instances_write(
        self,
        cache: set[tuple[str, str]],
        view_by_read_class: dict[type[DomainModelCore], dm.ViewId] | None,
        write_none: bool = False,
    ) -> ResourcesWrite:
        raise NotImplementedError()

    @model_validator(mode="before")
    def create_external_id_if_factory(cls, data: Any) -> Any:
        if isinstance(data, dict) and cls.external_id_factory is not None:
            data["external_id"] = cls.external_id_factory(cls, data)
        return data

    @classmethod
    def from_instance(cls: type[T_DomainModelWrite], instance: InstanceApply) -> T_DomainModelWrite:
        data = instance.dump(camel_case=False)
        data.pop("instance_type", None)
        node_type = data.pop("type", None)
        space = data.pop("space")
        external_id = data.pop("external_id")
        sources = data.pop("sources", [])
        properties = {}
        for source in sources:
            for prop_name, prop_value in source["properties"].items():
                if isinstance(prop_value, dict) and "externalId" in prop_value and "space" in prop_value:
                    if prop_value["space"] == DEFAULT_INSTANCE_SPACE:
                        properties[prop_name] = prop_value["externalId"]
                    else:
                        properties[prop_name] = dm.NodeId(
                            space=prop_value["space"], external_id=prop_value["externalId"]
                        )
                else:
                    properties[prop_name] = prop_value
        return cls(
            space=space, external_id=external_id, node_type=node_type, data_record=DataRecordWrite(**data), **properties
        )


T_DomainModelWrite = TypeVar("T_DomainModelWrite", bound=DomainModelWrite)


class CoreList(UserList, Generic[T_DomainModelCore]):
    _INSTANCE: type[T_DomainModelCore]
    _PARENT_CLASS: type[DomainModelCore]

    def __init__(self, nodes: Collection[T_DomainModelCore] = None):
        super().__init__(nodes or [])

    # The dunder implementations are to get proper type hints
    def __iter__(self) -> Iterator[T_DomainModelCore]:
        return super().__iter__()

    @overload
    def __getitem__(self, item: int) -> T_DomainModelCore: ...

    @overload
    def __getitem__(self: type[T_DomainModelList], item: slice) -> T_DomainModelList: ...

    def __getitem__(self, item: int | slice) -> T_DomainModelCore | T_DomainModelList:
        if isinstance(item, slice):
            return self.__class__(self.data[item])
        elif isinstance(item, int):
            return self.data[item]
        else:
            raise TypeError(f"Expected int or slice, got {type(item)}")

    def dump(self) -> list[dict[str, Any]]:
        return [node.model_dump() for node in self.data]

    def as_external_ids(self) -> list[str]:
        return [node.external_id for node in self.data]

    def to_pandas(self) -> pd.DataFrame:
        """
        Convert the list of nodes to a pandas.DataFrame.

        Returns:
            A pandas.DataFrame with the nodes as rows.
        """
        df = pd.DataFrame(self.dump())
        if df.empty:
            df = pd.DataFrame(columns=self._INSTANCE.model_fields)
        # Reorder columns to have the most relevant first
        id_columns = ["space", "external_id"]
        end_columns = ["node_type", "data_record"]
        fixed_columns = set(id_columns + end_columns)
        columns = (
            id_columns + [col for col in df if col not in fixed_columns] + [col for col in end_columns if col in df]
        )
        return df[columns]

    def _repr_html_(self) -> str:
        return self.to_pandas()._repr_html_()  # type: ignore[operator]


class DomainModelList(CoreList[T_DomainModelCore]):
    _PARENT_CLASS = DomainModel

    def __init__(self, nodes: Collection[T_DomainModelCore] = None):
        super().__init__(nodes or [])

    @property
    def data_records(self) -> DataRecordList:
        return DataRecordList([node.data_record for node in self])

    def as_node_ids(self) -> list[dm.NodeId]:
        return [dm.NodeId(space=node.space, external_id=node.external_id) for node in self]


T_DomainModelList = TypeVar("T_DomainModelList", bound=DomainModelList, covariant=True)


class DomainModelWriteList(DomainModelList[T_DomainModelWrite]):
    _PARENT_CLASS = DomainModelWrite

    @property
    def data_records(self) -> DataRecordWriteList:
        return DataRecordWriteList([node.data_record for node in self])

    def to_instances_write(
        self, view_by_read_class: dict[type[DomainModelCore], dm.ViewId] | None = None, write_none: bool = False
    ) -> ResourcesWrite:
        cache: set[tuple[str, str]] = set()
        domains = ResourcesWrite()
        for node in self:
            result = node._to_instances_write(cache, view_by_read_class, write_none)
            domains.extend(result)
        return domains

    def to_instances_apply(
        self, view_by_read_class: dict[type[DomainModelCore], dm.ViewId] | None = None, write_none: bool = False
    ) -> ResourcesWrite:
        warnings.warn(
            "to_instances_apply is deprecated and will be removed in v1.0. Use to_instances_write instead.",
            UserWarning,
            stacklevel=2,
        )
        return self.to_instances_write(view_by_read_class, write_none)


T_DomainModelWriteList = TypeVar("T_DomainModelWriteList", bound=DomainModelWriteList, covariant=True)


class DomainRelation(DomainModelCore):
    edge_type: dm.DirectRelationReference
    start_node: dm.DirectRelationReference
    data_record: DataRecord

    @property
    def data_records(self) -> DataRecordList:
        return DataRecordList([node.data_record for node in self])

    def as_id(self) -> dm.EdgeId:
        return dm.EdgeId(space=self.space, external_id=self.external_id)

    @classmethod
    def from_instance(cls: type[T_DomainModel], instance: Instance) -> T_DomainModel:
        data = instance.dump(camel_case=False)
        data.pop("instance_type", None)
        edge_type = data.pop("type", None)
        start_node = data.pop("start_node")
        end_node = data.pop("end_node")
        space = data.pop("space")
        external_id = data.pop("external_id")
        return cls(
            space=space,
            external_id=external_id,
            data_record=DataRecord(**data),
            edge_type=edge_type,
            start_node=start_node,
            end_node=end_node,
            **unpack_properties(instance.properties),
        )


T_DomainRelation = TypeVar("T_DomainRelation", bound=DomainRelation)


def default_edge_external_id_factory(
    start_node: DomainModelWrite | str, end_node: DomainModelWrite | str, edge_type: dm.DirectRelationReference
) -> str:
    start = start_node if isinstance(start_node, str) else start_node.external_id
    end = end_node if isinstance(end_node, str) else end_node.external_id
    return f"{start}:{end}"


class DomainRelationWrite(BaseModel, extra=Extra.forbid, populate_by_name=True):
    external_id_factory: ClassVar[
        Callable[[Union[DomainModelWrite, str], Union[DomainModelWrite, str], dm.DirectRelationReference], str]
    ] = default_edge_external_id_factory
    data_record: DataRecordWrite = Field(default_factory=DataRecordWrite)
    external_id: Optional[str] = Field(None, min_length=1, max_length=255)

    @property
    def data_records(self) -> DataRecordWriteList:
        return DataRecordWriteList([node.data_record for node in self])

    @abstractmethod
    def _to_instances_write(
        self,
        cache: set[tuple[str, str]],
        start_node: DomainModelWrite,
        edge_type: dm.DirectRelationReference,
        view_by_read_class: dict[type[DomainModelCore], dm.ViewId] | None,
        write_none: bool = False,
    ) -> ResourcesWrite:
        raise NotImplementedError()

    @classmethod
    def create_edge(
        cls, start_node: DomainModelWrite | str, end_node: DomainModelWrite | str, edge_type: dm.DirectRelationReference
    ) -> dm.EdgeApply:
        if isinstance(start_node, DomainModelWrite):
            space = start_node.space
        elif isinstance(start_node, DomainModelWrite):
            space = start_node.space
        else:
            raise TypeError(f"Either pass in a start or end node of type {DomainRelationWrite.__name__}")

        if isinstance(end_node, str):
            end_ref = dm.DirectRelationReference(space, end_node)
        elif isinstance(end_node, DomainModelWrite):
            end_ref = end_node.as_direct_reference()
        else:
            raise TypeError(f"Expected str or subclass of {DomainRelationWrite.__name__}, got {type(end_node)}")

        if isinstance(start_node, str):
            start_ref = dm.DirectRelationReference(space, start_node)
        elif isinstance(start_node, DomainModelWrite):
            start_ref = start_node.as_direct_reference()
        else:
            raise TypeError(f"Expected str or subclass of {DomainRelationWrite.__name__}, got {type(start_node)}")

        return dm.EdgeApply(
            space=space,
            external_id=cls.external_id_factory(start_node, end_node, edge_type),
            type=edge_type,
            start_node=start_ref,
            end_node=end_ref,
        )

    @classmethod
    def from_edge_to_resources(
        cls,
        cache: set[tuple[str, str]],
        start_node: DomainModelWrite | str,
        end_node: DomainModelWrite | str,
        edge_type: dm.DirectRelationReference,
        view_by_read_class: dict[type[DomainModelCore], dm.ViewId] | None = None,
        write_none: bool = False,
    ) -> ResourcesWrite:
        resources = ResourcesWrite()
        edge = DomainRelationWrite.create_edge(start_node, end_node, edge_type)
        if (edge.space, edge.external_id) in cache:
            return resources
        resources.edges.append(edge)
        cache.add((edge.space, edge.external_id))

        if isinstance(end_node, DomainModelWrite):
            other_resources = end_node._to_instances_write(cache, view_by_read_class, write_none)
            resources.extend(other_resources)
        if isinstance(start_node, DomainModelWrite):
            other_resources = start_node._to_instances_write(cache, view_by_read_class, write_none)
            resources.extend(other_resources)

        return resources


T_DomainRelationWrite = TypeVar("T_DomainRelationWrite", bound=DomainRelationWrite)


class DomainRelationList(CoreList[T_DomainRelation]):
    _PARENT_CLASS = DomainRelation

    def as_edge_ids(self) -> list[dm.EdgeId]:
        return [edge.as_id() for edge in self]


T_DomainRelationList = TypeVar("T_DomainRelationList", bound=DomainRelationList)


def unpack_properties(properties: Properties) -> Mapping[str, PropertyValue]:
    unpacked: dict[str, PropertyValue] = {}
    for view_properties in properties.values():
        for prop_name, prop_value in view_properties.items():
            if isinstance(prop_value, dict) and "externalId" in prop_value and "space" in prop_value:
                if prop_value["space"] == DEFAULT_INSTANCE_SPACE:
                    unpacked[prop_name] = prop_value["externalId"]
                else:
                    unpacked[prop_name] = dm.NodeId(space=prop_value["space"], external_id=prop_value["externalId"])
            else:
                unpacked[prop_name] = prop_value
    return unpacked



from __future__ import annotations

import warnings
from typing import TYPE_CHECKING, Any, Literal, Optional, Union

from cognite.client import data_modeling as dm
from pydantic import Field

from ._core import (
    DEFAULT_INSTANCE_SPACE,
    DataRecordWrite,
    DomainModel,
    DomainModelCore,
    DomainModelWrite,
    DomainModelWriteList,
    DomainModelList,
    DomainRelationWrite,
    ResourcesWrite,
)

if TYPE_CHECKING:
    from ._alert import Alert, AlertWrite
    from ._bid_method_afrr import BidMethodAFRR, BidMethodAFRRWrite
    from ._bid_row import BidRow, BidRowWrite


__all__ = [
    "BidRow",
    "BidRowWrite",
    "BidRowApply",
    "BidRowList",
    "BidRowWriteList",
    "BidRowApplyList",
    "BidRowFields",
    "BidRowTextFields",
]


BidRowTextFields = Literal["product", "exclusive_group_id", "asset_type", "asset_id"]
BidRowFields = Literal[
    "price",
    "quantity_per_hour",
    "product",
    "is_divisible",
    "min_quantity",
    "is_block",
    "exclusive_group_id",
    "asset_type",
    "asset_id",
]

_BIDROW_PROPERTIES_BY_FIELD = {
    "price": "price",
    "quantity_per_hour": "quantityPerHour",
    "product": "product",
    "is_divisible": "isDivisible",
    "min_quantity": "minQuantity",
    "is_block": "isBlock",
    "exclusive_group_id": "exclusiveGroupId",
    "asset_type": "assetType",
    "asset_id": "assetId",
}


class BidRow(DomainModel):
    """This represents the reading version of bid row.

    It is used to when data is retrieved from CDF.

    Args:
        space: The space where the node is located.
        external_id: The external id of the bid row.
        data_record: The data record of the bid row node.
        price: Price in EUR/MW/h, rounded to nearest price step (0.1?)
        quantity_per_hour: The capacity offered, per hour, in MW, rounded to nearest step size (5?)
        product: The product field.
        is_divisible: The is divisible field.
        min_quantity: Min quantity, per hour. Only relevant for divisible Bids. The minimum capacity that must be accepted; this must be lower than capacityPerHour and is rounded to the nearest step (5 MW?)).
        is_block: Indication if the row is part of a Block bid. If true: quantityPerHour must have the same value for consecutive hours (and no breaks). Block bids must be accepted for all hours or none.
        exclusive_group_id: Other bids with the same ID are part of an exclusive group - only one of them can be accepted, and they must have the same direction (product). Not allowed for block bids.
        linked_bid: The linked bid must have the opposite direction (link means that both or none must be accepted). Should be bi-directional.
        asset_type: The asset type field.
        asset_id: The asset id field.
        method: The method field.
        alerts: An array of associated alerts.
    """

    space: str = DEFAULT_INSTANCE_SPACE
    node_type: Union[dm.DirectRelationReference, None] = dm.DirectRelationReference("sp_powerops_types", "BidRow")
    price: Optional[float] = None
    quantity_per_hour: Optional[list[float]] = Field(None, alias="quantityPerHour")
    product: Optional[str] = None
    is_divisible: Optional[bool] = Field(None, alias="isDivisible")
    min_quantity: Optional[list[float]] = Field(None, alias="minQuantity")
    is_block: Optional[bool] = Field(None, alias="isBlock")
    exclusive_group_id: Optional[str] = Field(None, alias="exclusiveGroupId")
    linked_bid: Union[BidRow, str, dm.NodeId, None] = Field(None, repr=False, alias="linkedBid")
    asset_type: Optional[str] = Field(None, alias="assetType")
    asset_id: Optional[str] = Field(None, alias="assetId")
    method: Union[BidMethodAFRR, str, dm.NodeId, None] = Field(None, repr=False)
    alerts: Union[list[Alert], list[str], None] = Field(default=None, repr=False)

    def as_write(self) -> BidRowWrite:
        """Convert this read version of bid row to the writing version."""
        return BidRowWrite(
            space=self.space,
            external_id=self.external_id,
            data_record=DataRecordWrite(existing_version=self.data_record.version),
            price=self.price,
            quantity_per_hour=self.quantity_per_hour,
            product=self.product,
            is_divisible=self.is_divisible,
            min_quantity=self.min_quantity,
            is_block=self.is_block,
            exclusive_group_id=self.exclusive_group_id,
            linked_bid=self.linked_bid.as_write() if isinstance(self.linked_bid, DomainModel) else self.linked_bid,
            asset_type=self.asset_type,
            asset_id=self.asset_id,
            method=self.method.as_write() if isinstance(self.method, DomainModel) else self.method,
            alerts=[alert.as_write() if isinstance(alert, DomainModel) else alert for alert in self.alerts or []],
        )

    def as_apply(self) -> BidRowWrite:
        """Convert this read version of bid row to the writing version."""
        warnings.warn(
            "as_apply is deprecated and will be removed in v1.0. Use as_write instead.",
            UserWarning,
            stacklevel=2,
        )
        return self.as_write()


class BidRowWrite(DomainModelWrite):
    """This represents the writing version of bid row.

    It is used to when data is sent to CDF.

    Args:
        space: The space where the node is located.
        external_id: The external id of the bid row.
        data_record: The data record of the bid row node.
        price: Price in EUR/MW/h, rounded to nearest price step (0.1?)
        quantity_per_hour: The capacity offered, per hour, in MW, rounded to nearest step size (5?)
        product: The product field.
        is_divisible: The is divisible field.
        min_quantity: Min quantity, per hour. Only relevant for divisible Bids. The minimum capacity that must be accepted; this must be lower than capacityPerHour and is rounded to the nearest step (5 MW?)).
        is_block: Indication if the row is part of a Block bid. If true: quantityPerHour must have the same value for consecutive hours (and no breaks). Block bids must be accepted for all hours or none.
        exclusive_group_id: Other bids with the same ID are part of an exclusive group - only one of them can be accepted, and they must have the same direction (product). Not allowed for block bids.
        linked_bid: The linked bid must have the opposite direction (link means that both or none must be accepted). Should be bi-directional.
        asset_type: The asset type field.
        asset_id: The asset id field.
        method: The method field.
        alerts: An array of associated alerts.
    """

    space: str = DEFAULT_INSTANCE_SPACE
    node_type: Union[dm.DirectRelationReference, None] = dm.DirectRelationReference("sp_powerops_types", "BidRow")
    price: Optional[float] = None
    quantity_per_hour: Optional[list[float]] = Field(None, alias="quantityPerHour")
    product: Optional[str] = None
    is_divisible: Optional[bool] = Field(None, alias="isDivisible")
    min_quantity: Optional[list[float]] = Field(None, alias="minQuantity")
    is_block: Optional[bool] = Field(None, alias="isBlock")
    exclusive_group_id: Optional[str] = Field(None, alias="exclusiveGroupId")
    linked_bid: Union[BidRowWrite, str, dm.NodeId, None] = Field(None, repr=False, alias="linkedBid")
    asset_type: Optional[str] = Field(None, alias="assetType")
    asset_id: Optional[str] = Field(None, alias="assetId")
    method: Union[BidMethodAFRRWrite, str, dm.NodeId, None] = Field(None, repr=False)
    alerts: Union[list[AlertWrite], list[str], None] = Field(default=None, repr=False)

    def _to_instances_write(
        self,
        cache: set[tuple[str, str]],
        view_by_read_class: dict[type[DomainModelCore], dm.ViewId] | None,
        write_none: bool = False,
    ) -> ResourcesWrite:
        resources = ResourcesWrite()
        if self.as_tuple_id() in cache:
            return resources

        write_view = (view_by_read_class or {}).get(BidRow, dm.ViewId("sp_powerops_models", "BidRow", "1"))

        properties: dict[str, Any] = {}

        if self.price is not None or write_none:
            properties["price"] = self.price

        if self.quantity_per_hour is not None or write_none:
            properties["quantityPerHour"] = self.quantity_per_hour

        if self.product is not None or write_none:
            properties["product"] = self.product

        if self.is_divisible is not None or write_none:
            properties["isDivisible"] = self.is_divisible

        if self.min_quantity is not None or write_none:
            properties["minQuantity"] = self.min_quantity

        if self.is_block is not None or write_none:
            properties["isBlock"] = self.is_block

        if self.exclusive_group_id is not None or write_none:
            properties["exclusiveGroupId"] = self.exclusive_group_id

        if self.linked_bid is not None:
            properties["linkedBid"] = {
                "space": self.space if isinstance(self.linked_bid, str) else self.linked_bid.space,
                "externalId": self.linked_bid if isinstance(self.linked_bid, str) else self.linked_bid.external_id,
            }

        if self.asset_type is not None or write_none:
            properties["assetType"] = self.asset_type

        if self.asset_id is not None or write_none:
            properties["assetId"] = self.asset_id

        if self.method is not None:
            properties["method"] = {
                "space": self.space if isinstance(self.method, str) else self.method.space,
                "externalId": self.method if isinstance(self.method, str) else self.method.external_id,
            }

        if properties:
            this_node = dm.NodeApply(
                space=self.space,
                external_id=self.external_id,
                existing_version=self.data_record.existing_version,
                type=self.node_type,
                sources=[
                    dm.NodeOrEdgeData(
                        source=write_view,
                        properties=properties,
                    )
                ],
            )
            resources.nodes.append(this_node)
            cache.add(self.as_tuple_id())

        edge_type = dm.DirectRelationReference("sp_powerops_types", "calculationIssue")
        for alert in self.alerts or []:
            other_resources = DomainRelationWrite.from_edge_to_resources(
                cache, start_node=self, end_node=alert, edge_type=edge_type, view_by_read_class=view_by_read_class
            )
            resources.extend(other_resources)

        if isinstance(self.linked_bid, DomainModelWrite):
            other_resources = self.linked_bid._to_instances_write(cache, view_by_read_class)
            resources.extend(other_resources)

        if isinstance(self.method, DomainModelWrite):
            other_resources = self.method._to_instances_write(cache, view_by_read_class)
            resources.extend(other_resources)

        return resources


class BidRowApply(BidRowWrite):
    def __new__(cls, *args, **kwargs) -> BidRowApply:
        warnings.warn(
            "BidRowApply is deprecated and will be removed in v1.0. Use BidRowWrite instead."
            "The motivation for this change is that Write is a more descriptive name for the writing version of the"
            "BidRow.",
            UserWarning,
            stacklevel=2,
        )
        return super().__new__(cls)


class BidRowList(DomainModelList[BidRow]):
    """List of bid rows in the read version."""

    _INSTANCE = BidRow

    def as_write(self) -> BidRowWriteList:
        """Convert these read versions of bid row to the writing versions."""
        return BidRowWriteList([node.as_write() for node in self.data])

    def as_apply(self) -> BidRowWriteList:
        """Convert these read versions of primitive nullable to the writing versions."""
        warnings.warn(
            "as_apply is deprecated and will be removed in v1.0. Use as_write instead.",
            UserWarning,
            stacklevel=2,
        )
        return self.as_write()


class BidRowWriteList(DomainModelWriteList[BidRowWrite]):
    """List of bid rows in the writing version."""

    _INSTANCE = BidRowWrite


class BidRowApplyList(BidRowWriteList): ...


def _create_bid_row_filter(
    view_id: dm.ViewId,
    min_price: float | None = None,
    max_price: float | None = None,
    product: str | list[str] | None = None,
    product_prefix: str | None = None,
    is_divisible: bool | None = None,
    is_block: bool | None = None,
    exclusive_group_id: str | list[str] | None = None,
    exclusive_group_id_prefix: str | None = None,
    linked_bid: str | tuple[str, str] | list[str] | list[tuple[str, str]] | None = None,
    asset_type: str | list[str] | None = None,
    asset_type_prefix: str | None = None,
    asset_id: str | list[str] | None = None,
    asset_id_prefix: str | None = None,
    method: str | tuple[str, str] | list[str] | list[tuple[str, str]] | None = None,
    external_id_prefix: str | None = None,
    space: str | list[str] | None = None,
    filter: dm.Filter | None = None,
) -> dm.Filter | None:
    filters = []
    if min_price is not None or max_price is not None:
        filters.append(dm.filters.Range(view_id.as_property_ref("price"), gte=min_price, lte=max_price))
    if isinstance(product, str):
        filters.append(dm.filters.Equals(view_id.as_property_ref("product"), value=product))
    if product and isinstance(product, list):
        filters.append(dm.filters.In(view_id.as_property_ref("product"), values=product))
    if product_prefix is not None:
        filters.append(dm.filters.Prefix(view_id.as_property_ref("product"), value=product_prefix))
    if isinstance(is_divisible, bool):
        filters.append(dm.filters.Equals(view_id.as_property_ref("isDivisible"), value=is_divisible))
    if isinstance(is_block, bool):
        filters.append(dm.filters.Equals(view_id.as_property_ref("isBlock"), value=is_block))
    if isinstance(exclusive_group_id, str):
        filters.append(dm.filters.Equals(view_id.as_property_ref("exclusiveGroupId"), value=exclusive_group_id))
    if exclusive_group_id and isinstance(exclusive_group_id, list):
        filters.append(dm.filters.In(view_id.as_property_ref("exclusiveGroupId"), values=exclusive_group_id))
    if exclusive_group_id_prefix is not None:
        filters.append(dm.filters.Prefix(view_id.as_property_ref("exclusiveGroupId"), value=exclusive_group_id_prefix))
    if linked_bid and isinstance(linked_bid, str):
        filters.append(
            dm.filters.Equals(
                view_id.as_property_ref("linkedBid"), value={"space": DEFAULT_INSTANCE_SPACE, "externalId": linked_bid}
            )
        )
    if linked_bid and isinstance(linked_bid, tuple):
        filters.append(
            dm.filters.Equals(
                view_id.as_property_ref("linkedBid"), value={"space": linked_bid[0], "externalId": linked_bid[1]}
            )
        )
    if linked_bid and isinstance(linked_bid, list) and isinstance(linked_bid[0], str):
        filters.append(
            dm.filters.In(
                view_id.as_property_ref("linkedBid"),
                values=[{"space": DEFAULT_INSTANCE_SPACE, "externalId": item} for item in linked_bid],
            )
        )
    if linked_bid and isinstance(linked_bid, list) and isinstance(linked_bid[0], tuple):
        filters.append(
            dm.filters.In(
                view_id.as_property_ref("linkedBid"),
                values=[{"space": item[0], "externalId": item[1]} for item in linked_bid],
            )
        )
    if isinstance(asset_type, str):
        filters.append(dm.filters.Equals(view_id.as_property_ref("assetType"), value=asset_type))
    if asset_type and isinstance(asset_type, list):
        filters.append(dm.filters.In(view_id.as_property_ref("assetType"), values=asset_type))
    if asset_type_prefix is not None:
        filters.append(dm.filters.Prefix(view_id.as_property_ref("assetType"), value=asset_type_prefix))
    if isinstance(asset_id, str):
        filters.append(dm.filters.Equals(view_id.as_property_ref("assetId"), value=asset_id))
    if asset_id and isinstance(asset_id, list):
        filters.append(dm.filters.In(view_id.as_property_ref("assetId"), values=asset_id))
    if asset_id_prefix is not None:
        filters.append(dm.filters.Prefix(view_id.as_property_ref("assetId"), value=asset_id_prefix))
    if method and isinstance(method, str):
        filters.append(
            dm.filters.Equals(
                view_id.as_property_ref("method"), value={"space": DEFAULT_INSTANCE_SPACE, "externalId": method}
            )
        )
    if method and isinstance(method, tuple):
        filters.append(
            dm.filters.Equals(view_id.as_property_ref("method"), value={"space": method[0], "externalId": method[1]})
        )
    if method and isinstance(method, list) and isinstance(method[0], str):
        filters.append(
            dm.filters.In(
                view_id.as_property_ref("method"),
                values=[{"space": DEFAULT_INSTANCE_SPACE, "externalId": item} for item in method],
            )
        )
    if method and isinstance(method, list) and isinstance(method[0], tuple):
        filters.append(
            dm.filters.In(
                view_id.as_property_ref("method"), values=[{"space": item[0], "externalId": item[1]} for item in method]
            )
        )
    if external_id_prefix is not None:
        filters.append(dm.filters.Prefix(["node", "externalId"], value=external_id_prefix))
    if isinstance(space, str):
        filters.append(dm.filters.Equals(["node", "space"], value=space))
    if space and isinstance(space, list):
        filters.append(dm.filters.In(["node", "space"], values=space))
    if filter:
        filters.append(filter)
    return dm.filters.And(*filters) if filters else None
